from __future__ import annotations

import json
import os
import time
from datetime import datetime, timezone
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Dict, List, Optional
from urllib.parse import urljoin, urlparse

import requests
import socketio
from dateutil.tz import tzlocal
from packaging import version
from pydantic import parse_obj_as
from rich.status import Status
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from servicefoundry.io.output_callback import OutputCallBack
from servicefoundry.lib.auth.servicefoundry_session import ServiceFoundrySession
from servicefoundry.lib.clients.utils import request_handling
from servicefoundry.lib.const import API_SERVER_RELATIVE_PATH, VERSION_PREFIX
from servicefoundry.lib.model.entity import (
    Application,
    CreateDockerRepositoryResponse,
    Deployment,
    DockerRegistryCredentials,
    TenantInfo,
    Token,
    TriggerJobResult,
    Workspace,
    WorkspaceResources,
)
from servicefoundry.lib.win32 import allow_interrupt
from servicefoundry.logger import logger
from servicefoundry.v2.lib.models import (
    AppDeploymentStatusResponse,
    ApplicationFqnResponse,
    BuildResponse,
    DeploymentFqnResponse,
)
from servicefoundry.version import __version__

DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE = "DEPLOYMENT_LOGS"
BUILD_LOGS_SUBSCRIBE_MESSAGE = "logs"

if TYPE_CHECKING:
    from servicefoundry.auto_gen.models import Application


def _upload_packaged_code(metadata, package_file):
    file_size = os.stat(package_file).st_size
    with open(package_file, "rb") as file_to_upload:
        with tqdm(
            total=file_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            desc="Uploading package",
        ) as progress_bar:
            wrapped_file = CallbackIOWrapper(
                progress_bar.update, file_to_upload, "read"
            )
            http_response = requests.put(metadata["url"], data=wrapped_file)

            if http_response.status_code not in [204, 201, 200]:
                raise RuntimeError(f"Failed to upload code {http_response.content}")


class ServiceFoundryServiceClient:
    def __init__(self, init_session: bool = True, base_url: Optional[str] = None):
        self._session: Optional[ServiceFoundrySession] = None
        if init_session:
            if base_url:
                logger.warning("Passed base url %r will be ignored", base_url)
            self._session = ServiceFoundrySession()
            base_url = self._session.base_url
        elif not base_url:
            raise Exception("Neither session, not base_url provided")

        self._base_url = base_url.strip("/")
        self._api_server_url = f"{self._base_url}/{API_SERVER_RELATIVE_PATH}"
        if __version__ != "0.0.0":
            # "0.0.0" indicates dev version
            min_cli_version = self._get_min_cli_version_requirement()
            if version.parse(__version__) < version.parse(min_cli_version):
                raise Exception(
                    "You are using an outdated version of `servicefoundry`.\n"
                    f"Execute `pip install servicefoundry>={min_cli_version}` to install the supported version.",
                )
        else:
            logger.debug("Ignoring minimum cli version check")

    @property
    def base_url(self) -> str:
        return self._base_url

    def get_tenant_info(self) -> TenantInfo:
        res = requests.get(
            url=f"{self._api_server_url}/v1/tenant-id",
            params={"hostName": urlparse(self._api_server_url).netloc},
        )
        res = request_handling(res)
        return TenantInfo.parse_obj(res)

    @lru_cache(maxsize=3)
    def _get_min_cli_version_requirement(self) -> str:
        url = f"{self._api_server_url}/v1/min-cli-version"
        res = requests.get(url)
        res = request_handling(res)
        return res["minVersion"]

    def _get_header(self):
        if not self._session:
            return {}
        return {"Authorization": f"Bearer {self._session.access_token}"}

    def get_id_from_fqn(self, fqn_type: str, fqn: str):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/{fqn_type}"
        res = requests.get(url, headers=self._get_header(), params={"fqn": fqn})
        return request_handling(res)

    def list_workspace(self):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def list_workspaces(
        self,
        cluster_id: Optional[str] = None,
        workspace_name: Optional[str] = None,
        workspace_fqn: Optional[str] = None,
    ) -> List[Workspace]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        params = {}
        if cluster_id:
            params["clusterId"] = cluster_id
        if workspace_name:
            params["workspaceName"] = workspace_name
        if workspace_fqn:
            params["workspaceFqn"] = workspace_fqn
        res = requests.get(url, params=params, headers=self._get_header())
        response = request_handling(res)
        return parse_obj_as(List[Workspace], response)

    def create_workspace(
        self,
        workspace_name: str,
        cluster_name: str,
        resources: WorkspaceResources,
    ) -> Workspace:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.post(
            url,
            json={
                "manifest": {
                    "cluster": cluster_name,
                    "name": workspace_name,
                    "resources": resources.dict(exclude_none=True),
                }
            },
            headers=self._get_header(),
        )
        res = request_handling(res)
        return Workspace.parse_obj(res)

    def remove_workspace(self, workspace_id, force=False) -> Workspace:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace/{workspace_id}"
        force = json.dumps(
            force
        )  # this dumb conversion is required because `params` just casts as str
        res = requests.delete(url, headers=self._get_header(), params={"force": force})
        response = request_handling(res)
        return Workspace.parse_obj(response["workspace"])

    def get_workspace_by_name(self, workspace_name, cluster_id):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.get(
            url,
            headers=self._get_header(),
            params={"name": workspace_name, "clusterId": cluster_id},
        )
        return request_handling(res)

    def get_workspace(self, workspace_id):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace/{workspace_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_workspace_by_fqn(self, workspace_fqn: str) -> List[Workspace]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.get(
            url,
            headers=self._get_header(),
            params={"fqn": workspace_fqn},
        )
        response = request_handling(res)
        return parse_obj_as(List[Workspace], response)

    def list_deployments(self, workspace_id: str = None):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment"
        params = {}
        if workspace_id:
            params["workspaceId"] = workspace_id
        res = requests.get(url=url, params=params, headers=self._get_header())
        return request_handling(res)

    def list_cluster(self):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/cluster"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_cluster(self, cluster_id):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/cluster/{cluster_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    def get_presigned_url(self, space_name, service_name, env):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment/code-upload-url"
        res = requests.post(
            url,
            json={
                "workspaceFqn": space_name,
                "serviceName": service_name,
                "stage": env,
            },
            headers=self._get_header(),
        )
        return request_handling(res)

    def upload_code_package(
        self, workspace_fqn: str, component_name: str, package_local_path: str
    ) -> str:
        http_response = self.get_presigned_url(
            space_name=workspace_fqn, service_name=component_name, env="default"
        )
        _upload_packaged_code(metadata=http_response, package_file=package_local_path)

        return http_response["uri"]

    def deploy_application(
        self, workspace_id: str, application: Application
    ) -> Deployment:
        data = {
            "workspaceId": workspace_id,
            "name": application.name,
            "manifest": application.dict(exclude_none=True),
        }
        logger.debug(json.dumps(data))
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment"
        deploy_response = requests.post(url, json=data, headers=self._get_header())
        response = request_handling(deploy_response)
        return Deployment.parse_obj(response["deployment"])

    def _get_log_print_line(self, log: dict):
        timestamp = int(log["time"]) / 1e6

        time_obj = datetime.fromtimestamp(timestamp / 1000.0, tz=timezone.utc)
        time_obj.replace(tzinfo=timezone.utc)
        local_time = time_obj.astimezone(tzlocal())
        local_time_str = local_time.isoformat()
        return f'[{local_time_str}] {log["log"].strip()}'

    def _tail_logs(
        self,
        tail_logs_url: str,
        query_dict: dict,
        # NOTE: Rather making this printer callback an argument,
        # we should have global printer callback
        # which will be initialized based on the running env (cli, lib, notebook)
        subscribe_message: str,
        socketio_path: str = "socket.io",
        callback=OutputCallBack(),
        wait=True,
    ):

        sio = socketio.Client(request_timeout=60)
        callback.print_line("Waiting for the task to start...")

        @sio.on(subscribe_message)
        def logs(data):
            try:
                _log = json.loads(data)
                callback.print_line(self._get_log_print_line(_log["body"]))
            except Exception:
                logger.exception(f"Error while parsing log line, {data!r}")

        def sio_disconnect_no_exception():
            try:
                sio.disconnect()
            except Exception:
                logger.exception("Error while disconnecting from socket connection")

        with allow_interrupt(sio_disconnect_no_exception):
            sio.connect(
                tail_logs_url,
                transports="websocket",
                headers=self._get_header(),
                socketio_path=socketio_path,
            )
            # TODO: We should have have a timeout here. `emit` does
            #   not support timeout. Explore `sio.call`.
            sio.emit(
                subscribe_message,
                json.dumps(query_dict),
            )
            if wait:
                sio.wait()

    def get_deployment(self, application_id: str, deployment_id: str) -> Deployment:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments/{deployment_id}"
        res = requests.get(url, headers=self._get_header())
        res = request_handling(res)
        return Deployment.parse_obj(res)

    def get_deployment_statuses(
        self, application_id: str, deployment_id: str
    ) -> List[AppDeploymentStatusResponse]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments/{deployment_id}/statuses"
        res = requests.get(url, headers=self._get_header())
        res = request_handling(res)
        return parse_obj_as(List[AppDeploymentStatusResponse], res)

    def get_deployment_build_response(
        self, application_id: str, deployment_id: str
    ) -> List[BuildResponse]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments/{deployment_id}/builds"
        res = requests.get(url, headers=self._get_header())
        res = request_handling(res)
        return parse_obj_as(List[BuildResponse], res)

    def _get_deployment_logs(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        start_ts_nano: Optional[int] = None,
        end_ts_nano: Optional[int] = None,
        limit: Optional[int] = None,
        num_logs_to_ignore: Optional[int] = None,
    ) -> List:
        get_logs_query = {"applicationId": application_id}
        if deployment_id:
            get_logs_query["deploymentId"] = deployment_id
        data = {"getLogsQuery": json.dumps(get_logs_query)}
        if start_ts_nano:
            data["startTs"] = str(start_ts_nano)
        if end_ts_nano:
            data["endTs"] = str(end_ts_nano)
        if limit:
            data["limit"] = str(limit)
        if num_logs_to_ignore:
            data["numLogsToIgnore"] = int(num_logs_to_ignore)

        url = f"{self._api_server_url}/{VERSION_PREFIX}/logs/{workspace_id}"
        res = requests.get(url=url, params=data, headers=self._get_header())
        res = request_handling(res)
        return list(res["logs"])

    def tail_build_logs(
        self,
        build_response: BuildResponse,
        callback=OutputCallBack(),
        wait: bool = True,
    ):
        tail_logs_obj = json.loads(build_response.tailLogsUrl)
        self._tail_logs(
            tail_logs_url=tail_logs_obj["uri"],
            socketio_path=tail_logs_obj["path"],
            query_dict={
                "pipelineRunName": build_response.name,
                "startTs": build_response.logsStartTs,
            },
            callback=callback,
            wait=wait,
            subscribe_message=BUILD_LOGS_SUBSCRIBE_MESSAGE,
        )

    def tail_logs_for_deployment(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        start_ts: int,
        limit: int,
        callback=OutputCallBack(),
        wait: bool = True,
    ):
        self._tail_logs(
            tail_logs_url=urljoin(
                self._api_server_url, f"/?type={DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE}"
            ),
            query_dict={
                "workspaceId": workspace_id,
                "startTs": str(int(start_ts * 1e6)),
                "limit": limit,
                "getLogsQuery": {
                    "applicationId": application_id,
                    "deploymentId": deployment_id,
                },
            },
            callback=callback,
            wait=wait,
            subscribe_message=DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE,
        )

    def poll_logs_for_deployment(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        start_ts: int,
        limit: int,
        poll_interval_seconds: int,
        callback=OutputCallBack(),
    ):
        start_ts_nano = int(start_ts * 1e6)

        with Status(status="Polling for logs") as spinner:
            num_logs_to_ignore = 0

            while True:
                logs_list = self._get_deployment_logs(
                    workspace_id=workspace_id,
                    application_id=application_id,
                    deployment_id=deployment_id,
                    start_ts_nano=start_ts_nano,
                    limit=limit,
                    num_logs_to_ignore=num_logs_to_ignore,
                )

                if len(logs_list) == 0:
                    logger.warning("Did not receive any logs")
                    time.sleep(poll_interval_seconds)
                    continue

                for log in logs_list:
                    callback.print_line(self._get_log_print_line(log))

                last_log_time = logs_list[-1]["time"]
                num_logs_to_ignore = 0
                for log in reversed(logs_list):
                    if log["time"] != last_log_time:
                        break
                    num_logs_to_ignore += 1

                start_ts_nano = int(last_log_time)
                spinner.update(status=f"Waiting for {poll_interval_seconds} secs.")
                time.sleep(poll_interval_seconds)

    def fetch_deployment_logs(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        start_ts: Optional[int],
        end_ts: Optional[int],
        limit: Optional[int],
        callback=OutputCallBack(),
    ):
        logs_list = self._get_deployment_logs(
            workspace_id=workspace_id,
            application_id=application_id,
            deployment_id=deployment_id,
            start_ts_nano=int(start_ts * 1e6),
            end_ts_nano=int(end_ts * 1e6),
            limit=limit,
        )
        for log in logs_list:
            callback.print_line(self._get_log_print_line(log))

    def fetch_build_logs(
        self,
        build_response: BuildResponse,
        callback=OutputCallBack(),
    ) -> None:
        url = build_response.getLogsUrl
        res = requests.get(url=url, headers=self._get_header())
        logs_list = request_handling(res)
        for log in logs_list["logs"]:
            # TODO: Have to establish a log line format that includes timestamp, level, message
            callback.print_line(self._get_log_print_line(log))

    def get_deployment_info_by_fqn(self, deployment_fqn: str) -> DeploymentFqnResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/deployment"
        res = requests.get(
            url, headers=self._get_header(), params={"fqn": deployment_fqn}
        )
        res = request_handling(res)
        return DeploymentFqnResponse.parse_obj(res)

    def get_application_info_by_fqn(
        self, application_fqn: str
    ) -> ApplicationFqnResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/app"
        res = requests.get(
            url, headers=self._get_header(), params={"fqn": application_fqn}
        )
        res = request_handling(res)
        return ApplicationFqnResponse.parse_obj(res)

    def remove_application(self, application_id: str):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}"
        res = requests.delete(url, headers=self._get_header())
        response = request_handling(res)
        # TODO: Add pydantic here.
        return response

    def get_application_info(self, application_id: str) -> Application:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}"
        res = requests.get(url, headers=self._get_header())
        response = request_handling(res)
        return Application.parse_obj(response)

    def process_infra(self, infra_manifest: Any) -> Any:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/infra"
        # url = "http://localhost:8080/manifests/infra"
        res = requests.post(
            url,
            headers=self._get_header(),
            json={"manifest": infra_manifest},
        )
        response = request_handling(res)
        return response

    def trigger_job(
        self,
        deployment_id: str,
        component_name: str,
        command: Optional[str] = None,
        params: Optional[Dict[str, str]] = None,
    ) -> TriggerJobResult:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/trigger"
        body = {
            "deploymentId": deployment_id,
            "componentName": component_name,
            "input": {},
        }
        if command:
            body["input"]["command"] = command
        if params:
            body["input"]["params"] = params
        res = requests.post(url, json=body, headers=self._get_header())
        response = request_handling(res)
        return TriggerJobResult.parse_obj(response)

    def get_docker_registry_creds(
        self, docker_registry_fqn: str
    ) -> DockerRegistryCredentials:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/docker-registry/creds"
        res = requests.get(
            url,
            headers=self._get_header(),
            params={
                "fqn": docker_registry_fqn,
            },
        )
        response = request_handling(res)
        return DockerRegistryCredentials.parse_obj(response)

    def create_repo_in_registry(
        self, docker_registry_fqn: str, workspace_fqn: str, application_name: str
    ) -> CreateDockerRepositoryResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/docker-registry/create-repo"
        res = requests.post(
            url,
            headers=self._get_header(),
            data={
                "fqn": docker_registry_fqn,
                "workspaceFqn": workspace_fqn,
                "applicationName": application_name,
            },
        )
        response = request_handling(res)
        return CreateDockerRepositoryResponse.parse_obj(response)

    def list_applications(
        self,
        application_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        application_name: Optional[str] = None,
    ) -> List[Application]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app"
        params = {}
        if application_id:
            params["applicationId"] = application_id
        if workspace_id:
            params["workspaceId"] = workspace_id
        if application_name:
            params["applicationName"] = application_name
        res = requests.get(url, params=params, headers=self._get_header())
        response = request_handling(res)
        return parse_obj_as(List[Application], response)

    def list_versions(
        self,
        application_id: str,
        deployment_version: Optional[int] = None,
        deployment_id: Optional[str] = None,
    ) -> List[Deployment]:
        url = (
            f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments"
        )
        params = {}
        if deployment_version:
            params["version"] = deployment_version
        if deployment_id:
            params["deploymentId"] = deployment_id
        res = requests.get(url, params=params, headers=self._get_header())
        response = request_handling(res)
        return parse_obj_as(List[Deployment], response)

    def get_token_from_api_key(self, api_key: str) -> Token:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/oauth/api-key/token"
        data = {"apiKey": api_key}
        res = requests.get(url, params=data)
        res = request_handling(res)
        return Token.parse_obj(res)



from __future__ import annotations

import os
import threading
from functools import lru_cache, wraps
from pathlib import Path
from typing import Optional

from filelock import FileLock, Timeout

from servicefoundry.lib.const import CREDENTIAL_FILEPATH
from servicefoundry.lib.model.entity import CredentialsFileContent
from servicefoundry.logger import logger


def _ensure_lock_taken(method):
    @wraps(method)
    def lock_guard(self, *method_args, **method_kwargs):
        if not self.lock_taken():
            raise Exception(
                "Trying to write to credential file without using with block"
            )
        return method(self, *method_args, **method_kwargs)

    return lock_guard


CRED_FILE_THREAD_LOCK = threading.RLock()


@lru_cache(maxsize=None)
def get_file_lock(lock_file_path: str) -> FileLock:
    return FileLock(lock_file_path)


class CredentialsFileManager:
    def __init__(
        self,
        credentials_file_path: Path = CREDENTIAL_FILEPATH,
        lock_timeout: float = 60.0,
    ) -> None:
        credentials_file_path = credentials_file_path.absolute()
        logger.debug("credential file path %r", credentials_file_path)

        credentials_lock_file_path = f"{credentials_file_path}.lock"
        logger.debug("credential lock file path %r", credentials_lock_file_path)

        self._credentials_file_path = credentials_file_path
        cred_file_dir = credentials_file_path.parent
        cred_file_dir.mkdir(exist_ok=True, parents=True)

        self._file_lock = get_file_lock(credentials_lock_file_path)
        self._lock_timeout = lock_timeout
        self._lock_owner: Optional[int] = None

    def __enter__(self) -> CredentialsFileManager:
        # The lock objects are recursive locks, which means that once acquired, they will not block on successive lock requests:
        lock_aquired = CRED_FILE_THREAD_LOCK.acquire(timeout=self._lock_timeout)
        if not lock_aquired:
            raise Exception(
                "Could not aquire CRED_FILE_THREAD_LOCK"
                f" in {self._lock_timeout} seconds"
            )
        try:
            self._file_lock.acquire(timeout=self._lock_timeout)
        except Timeout as ex:
            raise Exception(
                f"Failed to aquire lock on credential file within {self._lock_timeout} seconds.\n"
                "Is any other process trying to login?"
            ) from ex
        logger.debug("Acquired file and thread lock to access credential file")
        self._lock_owner = threading.get_ident()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._file_lock.release()
        CRED_FILE_THREAD_LOCK.release()
        logger.debug("Released file and thread lock to access credential file")
        self._lock_owner = None

    def lock_taken(self) -> bool:
        return self._lock_owner == threading.get_ident()

    @_ensure_lock_taken
    def read(self) -> CredentialsFileContent:
        try:
            return CredentialsFileContent.parse_file(self._credentials_file_path)
        except Exception as ex:
            raise Exception(
                "Error while reading the credentials file "
                f"{self._credentials_file_path}. Please login again "
                "using `sfy login --relogin` or `sfy.login(relogin=True)` function"
            ) from ex

    @_ensure_lock_taken
    def write(self, credentials_file_content: CredentialsFileContent) -> None:
        if not isinstance(credentials_file_content, CredentialsFileContent):
            raise Exception(
                "Only object of type `CredentialsFileContent` is allowed. "
                f"Got {type(credentials_file_content)}"
            )
        logger.debug("Updating the credential file content")
        with open(self._credentials_file_path, "w", encoding="utf8") as file:
            file.write(credentials_file_content.json())

    @_ensure_lock_taken
    def delete(self) -> bool:
        if not os.path.exists(self._credentials_file_path):
            return False
        os.remove(self._credentials_file_path)
        return True

    @_ensure_lock_taken
    def exists(self) -> bool:
        return self._credentials_file_path.exists()



import os
import time
from typing import Any, Callable, Generator, Optional

from servicefoundry.lib.const import DEFAULT_BASE_URL, HOST_ENV_NAME
from servicefoundry.lib.exceptions import BadRequestException


def request_handling(res):
    try:
        status_code = res.status_code
    except Exception as ex:
        raise Exception("Unknown error occurred. Couldn't get status code.") from ex
    if 200 <= status_code <= 299:
        if res.content == b"":
            return None
        return res.json()
    if 400 <= status_code <= 499:
        try:
            message = str(res.json()["message"])
        except Exception:
            message = res
        raise BadRequestException(res.status_code, message)
    if 500 <= status_code <= 599:
        raise Exception(res.content)


def poll_for_function(
    func: Callable, poll_after_secs: int = 5, *args, **kwargs
) -> Generator[Any, None, None]:
    while True:
        yield func(*args, **kwargs)
        time.sleep(poll_after_secs)


def resolve_base_url(host: Optional[str] = None) -> str:
    return host or os.getenv(HOST_ENV_NAME) or DEFAULT_BASE_URL



import json
import sys
from typing import List, Optional

from pydantic import BaseModel, constr

from servicefoundry.function_service.route import RouteGroups
from servicefoundry.logger import logger
from servicefoundry.v2.lib.patched_models import PythonBuild
from servicefoundry.version import __version__


class BuildConfig(BaseModel):
    python_version: constr(
        regex=r"^\d+(\.\d+){1,2}$"
    ) = f"{sys.version_info.major}.{sys.version_info.minor}"
    pip_packages: Optional[List[str]]
    requirements_path: Optional[str] = None

    def __init__(self, **data):
        pip_packages = data.get("pip_packages", [])
        # locally version == 0.0.0
        # pip_packages.append(f"servicefoundry=={__version__}")

        if __version__ in ("NA", "0.0.0"):
            # TODO (chiragjn): Any change to servicefoundry.function_service.__main__ is untestable!
            #   We need to vendor `servicefoundry.function_service` parts in the local source to be able to test.
            sfy_version = ">=0.7.0,<0.9.0"
            logger.info(
                "Could not detect servicefoundry version. Using %r", sfy_version
            )
        else:
            sfy_version = f"=={__version__}"

        pip_packages.append(f"servicefoundry{sfy_version}")
        data["pip_packages"] = pip_packages
        super().__init__(**data)

    def to_tfy_python_build_config(
        self, port: int, route_groups: RouteGroups
    ) -> PythonBuild:
        escaped_route_groups_json = json.dumps(route_groups.json())
        return PythonBuild(
            python_version=self.python_version,
            pip_packages=self.pip_packages,
            requirements_path=self.requirements_path,
            command=f"python -m servicefoundry.function_service run --port {port} --route-groups-json {escaped_route_groups_json}",
        )



import collections
import os
from typing import Any, Dict, Iterator, List

import docker
from docker.errors import APIError, BuildError
from pydantic import BaseModel
from rich.console import Console

from servicefoundry.lib.clients.shell_client import Shell
from servicefoundry.logger import logger

__all__ = [
    "build_docker_image",
    "push_docker_image",
    "pull_docker_image",
    "push_docker_image_with_latest_tag",
]

PROGRESS_PERCENTAGE_DELTA = 20


class LastShownProgress(BaseModel):
    class Config:
        allow_mutation = True

    status: str = "Unknown"
    percentage: float = -1000.0


def _handle_layer_pull_progress_log(
    log_dict: Dict[str, Any], layer_last_shown_progress
):
    should_log = False
    id_, progress_detail, status = (
        log_dict["id"],
        log_dict.get("progressDetail", {}),
        log_dict["status"],
    )
    line_parts = [f"{id_}:", f"{status}"]
    if status != layer_last_shown_progress[id_].status:
        # Print if status of the layer has changed e.g. Downloading -> Extracting
        should_log = True
        layer_last_shown_progress[id_].status = status
    if not progress_detail:
        # If there is no progress information, reset last shown percentage
        layer_last_shown_progress[id_].percentage = -1000.0
    else:
        completed = progress_detail.get("current")
        total = progress_detail.get("total")
        if completed is not None and total is not None:
            last = layer_last_shown_progress[id_].percentage
            current = float(completed) / float(total) * 100.0
            # Print if the stage is 100% complete or progress of >= PROGRESS_PERCENTAGE_DELTA% was made over last time
            if (
                current >= 100.0
                or current // PROGRESS_PERCENTAGE_DELTA
                > last // PROGRESS_PERCENTAGE_DELTA
            ):
                should_log = True
                line_parts.append(
                    f"{float(completed) / 1e6:.3f}MB/{float(total) / 1e6:.3f}MB ({current:.2f}%)"
                )
                layer_last_shown_progress[id_].percentage = current
    return line_parts if should_log else []


def _stream_docker_build_logs(logs: Iterator[Dict[str, Any]]):
    console = Console(color_system=None, markup=False, soft_wrap=True)
    layer_last_shown_progress = collections.defaultdict(LastShownProgress)
    for log_dict in logs:
        parts = []
        if "id" in log_dict and "status" in log_dict:
            # If this line is a potential progress bar line
            parts = _handle_layer_pull_progress_log(log_dict, layer_last_shown_progress)
        else:
            for value in log_dict.values():
                line = str(value)
                if line.strip():
                    parts.append(line.rstrip())
        if parts:
            console.print(" ".join(parts))


def _get_docker_client():
    try:
        return docker.from_env()
    except Exception as ex:
        raise Exception("Could not connect to Docker") from ex


def env_has_docker():
    try:
        _get_docker_client()
        return True
    except Exception as ex:
        return False


# this is required since push does throw an error if it
# fails - so we have to parse the response logs to catch the error
# the other option is to run `docker login` as a subprocess, but it's
# not recommended to provide password to subprocess
def catch_error_in_push(response: List[dict]):
    for line in response:
        if line.get("error") is not None:
            raise Exception(
                f'Failed to push to registry with message \'{line.get("error")}\''
            )


def build_docker_image(
    path: str, tag: str, platform: str, dockerfile: str, cache_from: List[str]
):
    use_depot = bool(os.environ.get("USE_DEPOT"))
    depot_project_id = os.environ.get("DEPOT_PROJECT_KEY")
    logger.info("Starting docker build...")
    if use_depot and depot_project_id:
        try:
            Shell().execute_shell_command(
                [
                    "depot",
                    "build",
                    "--project",
                    depot_project_id,
                    "-f",
                    dockerfile,
                    "-t",
                    tag,
                    path,
                    "--push",
                ]
            )
        except Exception as e:
            raise Exception("Error while building Docker image using Depot") from e
    else:
        try:
            # TODO (chiragjn): Maybe consider using client.images.build
            build_logs = _get_docker_client().api.build(
                decode=True,
                path=path,
                tag=tag,
                platform=platform,
                dockerfile=dockerfile,
                cache_from=cache_from,
            )
            _stream_docker_build_logs(build_logs)
        except (BuildError, APIError) as e:
            raise Exception("Error while building Docker image") from e


def push_docker_image(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    client = _get_docker_client()
    auth_config = {"username": docker_login_username, "password": docker_login_password}

    logger.info(f"Pushing {image_uri}")
    response = client.images.push(
        repository=image_uri, auth_config=auth_config, decode=True, stream=True
    )
    catch_error_in_push(response=response)


def push_docker_image_with_latest_tag(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    client = _get_docker_client()
    auth_config = {"username": docker_login_username, "password": docker_login_password}

    repository_without_tag, _ = image_uri.rsplit(":", 1)
    image = client.images.get(image_uri)
    image.tag(repository=repository_without_tag, tag="latest")

    logger.info(f"Pushing {repository_without_tag}:latest")
    response = client.images.push(
        repository=repository_without_tag,
        tag="latest",
        auth_config=auth_config,
        decode=True,
        stream=True,
    )
    catch_error_in_push(response=response)


def pull_docker_image(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    auth_config = {"username": docker_login_username, "password": docker_login_password}
    logger.info(f"Pulling cache image {image_uri}")
    _get_docker_client().images.pull(repository=image_uri, auth_config=auth_config)



# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook


def is_notebook():
    try:
        # noinspection PyUnresolvedReferences
        shell = get_ipython().__class__
        if shell.__name__ == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell.__name__ == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        elif "google.colab" in str(shell):
            return True  # google colab notebook
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def get_default_callback():
    if is_notebook():
        from servicefoundry.io.notebook.io.notebook_callback import (
            NotebookOutputCallBack,
        )

        return NotebookOutputCallBack()
    else:
        from servicefoundry.io.rich_output_callback import RichOutputCallBack

        return RichOutputCallBack()



from rich.console import Console

# TODO: this file should not exist here, it is being imported even outside cli
#       moreover there is RichOutputCallBack now which can be used instead of console directly
console = Console()



from typing import Optional

from servicefoundry.io.notebook.notebook_util import get_default_callback
from servicefoundry.lib.session import login as session_login


def login(
    host: Optional[str] = None, api_key: Optional[str] = None, relogin: bool = False
):
    session_login(
        api_key=api_key, relogin=relogin, output_hook=get_default_callback(), host=host
    )



import inspect
import json
from typing import Any, Callable, Dict, List

from pydantic import BaseModel, Field, constr, validator

from servicefoundry.function_service.remote import RemoteClass
from servicefoundry.function_service.utils import (
    create_pydantic_model_from_function_signature,
    get_qual_name,
)
from servicefoundry.logger import logger


def validate_we_can_create_pydantic_model_of_func_args(func: Callable):
    qual_name = get_qual_name(func)
    try:
        create_pydantic_model_from_function_signature(func, get_qual_name(func))
    except Exception as ex:
        raise Exception(
            f"Unable to create a route for {qual_name!r}.\n"
            "Please ensure that in the function type signature, you have only used in-built\n"
            "types like `int`, `float`, `str`, `bool`, `typing.Dict`, `typing.List`, typing.Optional`.\n"
            "To temporarily resolve this error, you can remove the unsupported type signatures.\n"
        ) from ex


def path_pre_processor(path: str, prefix: str = "") -> str:
    path = path.strip("/")

    if not path:
        raise ValueError("path cannot be empty")

    prefix = prefix.strip("/")
    if not prefix:
        return f"/{path}"

    return f"/{prefix}/{path}"


class Route(BaseModel):
    function_name: str
    http_method: str
    path: constr(regex=r"^[A-Za-z0-9\-_/]+$")

    qual_name: str


class FunctionRoute(Route):
    module: str

    @classmethod
    def from_func(cls, func: Callable, path: str):
        validate_we_can_create_pydantic_model_of_func_args(func)
        return cls(
            function_name=func.__name__,
            http_method="POST",
            path=path_pre_processor(path or func.__name__),
            qual_name=get_qual_name(func),
            module=func.__module__,
        )


class ClassRoute(BaseModel):
    class_name: str
    init_kwargs: Dict[str, Any] = Field(default_factory=dict)
    module: str

    routes: List[Route] = Field(default_factory=list)

    @validator("init_kwargs")
    def init_kwargs_is_json_serializable(cls, v, values):
        try:
            json.dumps(v)
        except Exception as ex:
            class_name = values.get("class_name")
            raise ValueError(
                f"init_kwargs {v!r} of class {class_name!r} is not JSON serializable"
            ) from ex

        return v

    @classmethod
    def from_class(cls, remote_class: RemoteClass):
        routes = []
        methods = inspect.getmembers(remote_class.class_, predicate=inspect.isfunction)

        for method_name, method in methods:
            if method_name.startswith("_"):
                continue
            validate_we_can_create_pydantic_model_of_func_args(method)
            route = Route(
                function_name=method_name,
                http_method="POST",
                path=path_pre_processor(prefix=remote_class.name, path=method_name),
                qual_name=remote_class.get_qual_name(method),
            )
            routes.append(route)

        return cls(
            class_name=remote_class.class_.__name__,
            init_kwargs=remote_class.init_kwargs,
            routes=routes,
            module=remote_class.class_.__module__,
        )


class RouteGroups(BaseModel):
    functions: List[FunctionRoute] = Field(default_factory=list)
    classes: Dict[str, ClassRoute] = Field(default_factory=dict)

    def register_function(self, func, path):
        function_route = FunctionRoute.from_func(func=func, path=path)
        logger.info(
            "Function %r from module %r will be deployed on path '%s %s'.",
            function_route.function_name,
            function_route.module,
            function_route.http_method,
            function_route.path,
        )
        self.functions.append(function_route)

    def register_class(self, remote_class: RemoteClass):
        if remote_class.name in self.classes:
            raise ValueError(
                f"name {remote_class.name!r} is already used to register a class"
            )
        class_route = ClassRoute.from_class(remote_class)
        for route in class_route.routes:
            logger.info(
                "Method %r from `%s:%s` will be deployed on path '%s %s'.",
                route.function_name,
                class_route.class_name,
                remote_class.name,
                route.http_method,
                route.path,
            )
        self.classes[remote_class.name] = class_route



# generated by datamodel-codegen:
#   filename:  application.json
#   timestamp: 2023-03-24T12:37:18+00:00

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, confloat, conint, constr
from typing_extensions import Literal


class BasicAuthCreds(BaseModel):
    """
    +label=Username and password for service auth
    """

    type: Literal["basic_auth"]
    username: constr(regex=r"^[a-z0-9-@\.]{1,64}$") = Field(
        ...,
        description="+label=Username for service auth\n+message=Upto 64 lower case alphanumeric character long",
    )
    password: constr(regex=r"^[a-zA-Z0-9\.!@#$%^&*()_+=\-]{1,64}$") = Field(
        ...,
        description="+label=Password for service auth\n+message=Password should not be more than 64 characters",
    )


class BlueGreen(BaseModel):
    """
    +docs=This strategy brings up the new release completely before switching the complete load to the new release.
    This minimizes the time that two versions are serving traffic at the same time.
    +label=Blue Green strategy
    """

    type: Literal["blue_green"] = Field(..., description="+value=blue_green")
    enable_auto_promotion: bool = Field(
        False,
        description="+docs=Promote the new release to handle the complete traffic. A manual promotion would be needed if this is disabled\n+label=Auto-promotion",
    )
    auto_promotion_seconds: conint(ge=0) = Field(
        30,
        description="+docs=Promote the new release to handle the complete traffic after waiting for these many seconds\n+label=Auto-promotion seconds",
    )


class CPUUtilizationMetric(BaseModel):
    type: Literal["cpu_utilization"]
    value: conint(lt=100, gt=0) = Field(
        ...,
        description="+label=CPU utilization %\n+usage=Percentage of cpu request averaged over all replicas which the autoscaler should try to maintain",
    )


class CanaryStep(BaseModel):
    weight_percentage: conint(ge=0, le=100) = Field(
        ...,
        description="+docs=Percentage of total traffic to be shifted to the canary release.\nThe rest will continue to go to the existing deployment\n+label=Canary weight percentage\n+unit=%\n+placeholder=Weight",
    )
    pause_duration: conint(ge=0) = Field(
        30,
        description="+docs=Duration for which to pause the release. The release process will wait for these seconds before proceeding to the next step.\nIf this is not set, the step will pause indefinitely on this step\n+label=Pause duration\n+unit=seconds\n+placeholder=Duration",
    )


class DockerFileBuild(BaseModel):
    """
    +docs=Describes that we are using a dockerfile to build our image
    +label=Docker File (I already have Docker File)
    +icon=fa-brands fa-docker:#0db7ed
    """

    type: constr(regex=r"^dockerfile$") = Field(..., description="+value=dockerfile")
    dockerfile_path: str = Field(
        "./Dockerfile",
        description="+label=Path to Dockerfile\n+usage=The file path of the Dockerfile relative to project root path.",
    )
    build_context_path: str = Field(
        "./",
        description="+label=Path to build context\n+usage=Build context path for the Dockerfile relative to project root path.",
    )
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Entrypoint Override\n+usage=Override the command to run when the container starts\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )


class FileMount(BaseModel):
    mount_path: str = Field(
        ..., description="+usage=Path at which data is to be mounted"
    )
    is_base64_encoded: bool = Field(
        False, description="+usage=Set this flag if your data values are base64 encoded"
    )
    data: Dict[str, str] = Field(
        ...,
        description="+usage=Data to be mounted. Each key will be the filename and the value will be the file content. Files will be mounted at mount_path",
    )


class GPU(BaseModel):
    type: str = Field(
        ...,
        description="+label=Type\n+usage=Name of the Nvidia GPU. One of [K80, P4, P100, V100, T4, A10G, A100_40GB, A100_80GB]\nOne instance of the card contains the following amount of memory:\nK80: 12 GB\nP4: 8 GB\nP100: 16 GB\nV100: 16 GB\nT4: 16 GB\nA10G: 24 GB\nA100_40GB: 40GB\nA100_80GB: 80 GB",
    )
    count: conint(ge=1, le=16) = Field(
        1,
        description="+label=Count\n+usage=Count of GPUs to provide to the application\nNote the exact count and max count available for a given GPU type depends on cloud provider and cluster type.\nK80: 1-16 on AWS EKS, 1-8 on GCP GKE Standard, NA on GCP GKE Autopilot\nP4: NA on AWS EKS, 1-4 on GCP GKE Standard, NA on GCP GKE Autopilot\nP100: NA on AWS EKS, 1-4 on GCP GKE Standard, NA on GCP GKE Autopilot\nV100: 1-8 on AWS EKS, 1-8 on GCP GKE Standard, NA on GCP GKE Autopilot\nT4: 1-8 on AWS EKS, 1-4 on GCP GKE Standard, exactly 1 or 2 or 4 on GCP GKE Autopilot\nA10G: 1-8 on AWS EKS, NA on GCP GKE Standard, NA on GCP GKE Autopilot\nA100_40GB: 1-8 on AWS EKS, 1-16 on GCP GKE Standard, exactly 1 or 2 or 4 or 8 or 16 on GCP GKE Autopilot\nA100_80GB: 1-8 on AWS EKS, 1-8 on GCP GKE Standard, NA on GCP GKE Autopilot",
    )


class GitSource(BaseModel):
    """
    +docs=Describes that we are using code stored in a git repository to build our image
    +label=Git Source
    +icon=fa-solid fa-code-branch:black
    +sort=300
    """

    type: constr(regex=r"^git$") = Field(..., description="+value=git")
    repo_url: constr(
        regex=r"^(http(s?)://)(github\.com|(.+@)*bitbucket\.org|gitlab\.com).*$"
    ) = Field(
        ...,
        description="+label=Repo URL\n+usage=The repository URL.\n+sort=1\n+message=Needs to be a valid Github, Bitbucket or Gitlab link",
    )
    ref: str = Field(
        ..., description="+label=Commit SHA\n+usage=The commit SHA.\n+sort=2"
    )
    branch_name: Optional[str] = Field(
        None,
        description="+label=Branch Name\n+usage=Selecting branch will select latest commit SHA of the branch.\n+sort=3\n+ignore",
    )


class HttpProbe(BaseModel):
    """
    +docs=Describes the Instructions for assessing container health by executing an HTTP GET request.
    To learn more you can go [here](https://docs.truefoundry.com/docs/add-health-checks-to-deployments)
    +label=Instructions for assessing container health by executing an HTTP GET request.
    """

    type: constr(regex=r"^http$") = Field(..., description="+sort=1\n+value=http")
    path: str = Field(
        ...,
        description="+usage=The endpoint, relative to the port, to which the HTTP GET request should be directed.\n+sort=2",
    )
    port: conint(ge=0, le=65535) = Field(
        ...,
        description="+usage=The TCP socket within the container to which the HTTP GET request should be directed.\n+sort=3",
    )
    host: Optional[str] = Field(
        None,
        description="+sort=4\n+usage=Host name to connect to, defaults to the pod IP",
    )
    scheme: str = Field(
        "HTTP", description="+sort=5\n+usage=Scheme to use for connecting to the host"
    )


class HuggingfaceModelHub(BaseModel):
    """
    +docs=Huggingface model hub (deploy a model from HF model hub)
    +label=Huggingface model hub
    +usage=Deploy a model from HF model hub
    """

    type: constr(regex=r"^hf-model-hub$") = Field(
        ..., description="+value=hf-model-hub"
    )
    model_id: constr(regex=r"^[\w\-\._/]+$") = Field(
        ...,
        description="+label=Model repository name\n+usage=The name of the model that you want to deploy. Example:- `t5-small`, `philschmid/bart-large-cnn-samsum`.\n+placeholder=philschmid/bart-large-cnn-samsum",
    )
    pipeline: Optional[constr(regex=r"^[a-z0-9\-_]+$")] = Field(
        None,
        description="+label=Pipeline\n+usage=Pipeline to use for inference. If not set, we will try to infer the `task` name for the given model.\nExamples:- `summarization`, `text-generation`, `text-classification`, etc.\nYou can read more about HF pipelines here:- https://huggingface.co/docs/transformers/pipeline_tutorial\nYou can find a list of tasks here:- https://huggingface.co/docs/transformers/v4.24.0/en/main_classes/pipelines#transformers.pipeline.task\n+placeholder=summarization",
    )


class Image(BaseModel):
    """
    +docs=Describes that we are using a pre-built image stored in a Docker Image registry
    +label=Docker Image (Deploy an existing image)
    +icon=fa-brands fa-docker:#0db7ed
    """

    type: constr(regex=r"^image$") = Field(..., description="+value=image")
    image_uri: str = Field(
        ...,
        description="+label=Image URI\n+usage=The image URI. Specify the name of the image and the tag.\nIf the image is in Dockerhub, you can skip registry-url (for e.g. `tensorflow/tensorflow`).\nYou can use an image from a private registry using Advanced fields\n+placeholder=registry-url/account/image:version (e.g. docker.io/tensorflow/tensorflow)",
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Entrypoint Override\n+usage=Override the command to run when container starts.\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )


class LocalSource(BaseModel):
    """
    +docs=Describes that we are using code stored in a local developement environment to build our image
    +label=Local
    +icon=fa-folder:black
    +sort=100
    """

    type: constr(regex=r"^local$") = Field(..., description="+value=local")
    project_root_path: str = Field("./", description="+usage=Local project root path.")
    local_build: bool = Field(True, description="run docker build locally")


class Manual(BaseModel):
    """
    +docs=Describes that we are going to manually trigger our job.
    +label=Manual
    +usage=Trigger the job manually. [Docs](https://docs.truefoundry.com/docs/more-configurations)
    """

    type: constr(regex=r"^manual$") = Field(..., description="+value=manual")


class Param(BaseModel):
    name: constr(regex=r"^[a-z][a-z0-9\-_]{0,30}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the param\n+message=name can contain lower case alphabets, digits, underscore (_) and hypen (-). It can be 32 characters long, should start with an alphabet, and should end with either an alphabet or digit.",
    )
    description: Optional[constr(regex=r"^.{1,127}$")] = Field(
        None,
        description="+usage=Description of param\n+message=description cannot be longer than 127 characters",
    )
    default: Optional[constr(regex=r"^.{1,127}$")] = Field(
        None,
        description="+usage=Default value or placeholder\n+message=default value cannot be longer than 127 characters\n+label=Default value",
    )


class Protocol(str, Enum):
    """
    +usage=Protocol for the port.
    """

    TCP = "TCP"
    UDP = "UDP"


class AppProtocol(str, Enum):
    """
    +label=Application Protocol
    +usage=Application Protocol for the port.
    Select the application protocol used by your service. For most use cases, this should be `http`(HTTP/1.1).
    If you are running a gRPC server, select the `grpc` option.
    This is only applicable if `expose=true`.
    """

    http = "http"
    grpc = "grpc"
    tcp = "tcp"


class Port(BaseModel):
    """
    +docs=Describes the ports the service should be exposed to.
    """

    port: conint(ge=0, le=65535) = Field(
        80, description="+usage=Port number to expose."
    )
    protocol: Protocol = Field("TCP", description="+usage=Protocol for the port.")
    expose: bool = Field(True, description="+usage=Expose the port")
    host: Optional[
        constr(
            regex=r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
        )
    ] = Field(
        None,
        description="+usage=Host e.g. ai.example.com, app.truefoundry.com\n+message=Upto 253 characters, each part of host should be at most 63 characters long, can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts must be separated by periods (.)",
    )
    path: Optional[
        constr(regex=r"^(/([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-_\.]*[a-zA-Z0-9]))*/$")
    ] = Field(
        None,
        description="+usage=Path e.g. /v1/api/ml/, /v2/docs/\n+message=Should begin and end with a forward slash (/). Each part can can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts should be separated by forward slashes (/)",
    )
    rewrite_path_to: Optional[
        constr(regex=r"^(/([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-_\.]*[a-zA-Z0-9]))*/$")
    ] = Field(
        None,
        description="+label=Rewrite Path to\n+usage=Rewrite the path prefix to a different path.\nIf `path` is `/v1/api` and `rewrite_path_to` is `/api`. The URI in the HTTP request `http://0.0.0.0:8080/v1/api/houses` will be rewritten to `http://0.0.0.0:8080/api/houses` before the request is forwarded your service.\nDefaults to `/`.\nThis is only applicable if `path` is given.\n+message=Should begin and end with a forward slash (/). Each part can can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts should be separated by forward slashes (/)",
    )
    app_protocol: Optional[AppProtocol] = Field(
        None,
        description="+label=Application Protocol\n+usage=Application Protocol for the port.\nSelect the application protocol used by your service. For most use cases, this should be `http`(HTTP/1.1).\nIf you are running a gRPC server, select the `grpc` option.\nThis is only applicable if `expose=true`.",
    )
    auth: Optional[BasicAuthCreds] = None


class PythonBuild(BaseModel):
    """
    +docs=Describes that we are using python to build a container image with a specific python version and pip packages installed.
    +label=Python Code (I don't have Dockerfile)
    +icon=fa-brands fa-python:#306998
    """

    type: constr(regex=r"^tfy-python-buildpack$") = Field(
        ..., description="+value=tfy-python-buildpack"
    )
    python_version: constr(regex=r"^\d+(\.\d+){1,2}$") = Field(
        "3.9",
        description="+label=Python version\n+usage=Python version to run your application.\n+message=Please enter a valid Python version",
    )
    build_context_path: str = Field(
        "./",
        description="+label=Path to build context\n+usage=Build path relative to project root path.",
    )
    requirements_path: Optional[str] = Field(
        None,
        description="`Path to build context`\n+label=Path to requirements\n+usage=Path to `requirements.txt` relative to\n`Path to build context`",
    )
    pip_packages: Optional[List[str]] = Field(
        None,
        description="+label=Pip packages to install\n+usage=Define pip package requirements.",
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=Debian packages to install via apt.\n+usage=Define Debian packages to install via apt.\nE.g. `["ffmpeg", "libsm6", "libxext6"]',
    )
    command: Union[str, List[str]] = Field(
        ...,
        description="Command will be set as the Entrypoint of the generated\nimage.\n+label=Command\n+usage=Command to run when the container starts.\nCommand will be set as the Entrypoint of the generated image.\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )
    cuda_version: Optional[
        constr(
            regex=r"^((\d+\.\d+(\.\d+)?-cudnn\d+-(runtime|devel)-ubuntu\d+\.\d+)|9\.2-cudnn7|10\.0-cudnn7|10\.1-cudnn7|10\.1-cudnn8|10\.2-cudnn7|10\.2-cudnn8|11\.0-cudnn8|11\.1-cudnn8|11\.2-cudnn8|11\.3-cudnn8|11\.4-cudnn8|11\.5-cudnn8|11\.6-cudnn8|11\.7-cudnn8|11\.8-cudnn8)$"
        )
    ] = Field(
        None,
        description="+label=CUDA Version\n+usage=Version of CUDA Toolkit and CUDNN to install in the image\nThese combinations are based off of publically available docker images on docker hub\nYou can also specify a valid tag of the form {cuda_version_number}-cudnn{cudnn_version_number}-{runtime|devel}-ubuntu{ubuntu_version}\nRefer https://hub.docker.com/r/nvidia/cuda/tags for valid set of values\nNote: We use deadsnakes ubuntu ppa to add Python that currently supports only Ubuntu 18.04, 20.04 and 22.04",
    )


class RPSMetric(BaseModel):
    type: Literal["rps"]
    value: conint(ge=1) = Field(
        ...,
        description="+label=Requests per second\n+usage=Average request per second averaged over all replicas that autoscaler should try to maintain",
    )


class RemoteSource(BaseModel):
    """
    +docs=Describes that we are using code stored in a remote respository to build our image
    +label=S3
    +icon=fa-brands fa-aws:black
    +sort=200
    """

    type: constr(regex=r"^remote$") = Field(..., description="+value=remote")
    remote_uri: str = Field(
        ..., description="+docs=Remote repository URI\n+label=Remote URI"
    )


class Resources(BaseModel):
    """
    +docs=Describes the resource constraints for the application so that it can be deployed accordingly on the cluster
    To learn more you can go [here](https://docs.truefoundry.com/docs/resources)
    +icon=fa-microchip
    +label=Resources
    +usage=Set resource constraints for the application. [Docs](https://docs.truefoundry.com/docs/resources)
    """

    cpu_request: confloat(ge=0.001, le=16.0) = Field(
        0.2,
        description="+label=CPU Request\n+sort=1\n+usage=Requested CPU which determines the minimum cost incurred. The CPU usage can exceed the requested\namount, but not the value specified in the limit. 1 CPU means 1 CPU core. Fractional CPU can be requested\nlike `0.5` or `0.05`",
    )
    cpu_limit: confloat(ge=0.001, le=16.0) = Field(
        0.5,
        description="+label=CPU Limit\n+usage=CPU limit beyond which the usage cannot be exceeded. 1 CPU means 1 CPU core. Fractional CPU can be requested\nlike `0.5`. CPU limit should be >= cpu request.\n+sort=2",
    )
    memory_request: conint(ge=1, le=32000) = Field(
        200,
        description="+label=Memory Request\n+usage=Requested memory which determines the minimum cost incurred. The unit of memory is in megabytes(MB).\nSo 1 means 1 MB and 2000 means 2GB.\n+sort=3",
    )
    memory_limit: conint(ge=1, le=32000) = Field(
        500,
        description="+label=Memory Limit\n+usage=Memory limit after which the application will be killed with an OOM error. The unit of memory is\nin megabytes(MB). So 1 means 1 MB and 2000 means 2GB. MemoryLimit should be greater than memory request.\n+sort=4",
    )
    ephemeral_storage_request: conint(ge=1, le=500000) = Field(
        1000,
        description="+label=Storage Request\n+usage=Requested disk storage. The unit of memory is in megabytes(MB).\nThis is ephemeral storage and will be wiped out on pod restarts or eviction\n+sort=5",
    )
    ephemeral_storage_limit: conint(ge=1, le=500000) = Field(
        2000,
        description="+label=Storage Limit\n+usage=Disk storage limit. The unit of memory is in megabytes(MB). Exceeding this limit will result in eviction.\nIt should be greater than the request. This is ephemeral storage and will be wiped out on pod restarts or eviction\n+sort=6",
    )
    gpu: Optional[GPU] = None
    instance_family: Optional[List[str]] = Field(
        None,
        description="+label=Instance family\n+usage=Instance family of the underlying machine to use. Multiple instance families can be supplied.\nThe workload is guaranteed to be scheduled on one of them.",
    )
    shared_memory_size: Optional[conint(ge=64, le=32000)] = Field(
        None,
        description="+label=Shared Memory Size\n+usage=Define the shared memory requirements for your workload. Machine learning libraries like Pytorch can use Shared Memory\nfor inter-process communication. If you use this, we will mount a `tmpfs` backed volume at the `/dev/shm` directory.\nAny usage will also count against the workload's memory limit (`resources.memory_limit`) along with your workload's memory usage.\nIf the overall usage goes above `resources.memory_limit` the user process may get killed.\nShared Memory Size cannot be more than the defined Memory Limit for the workload.",
    )


class Rolling(BaseModel):
    """
    +docs=This strategy updates the pods in a rolling fashion such that a subset of the
    total pods are replaced with new version at one time.
    A commonly used strategy can be to have maxUnavailablePercentage close to 0 so that there
    is no downtime and keep the maxSurgePercentage to around 25%. If you are anyways running
    a large number of pods, the service can often tolerate a few pods going down - so you
    max maxUnavailablePercentage = 10 and maxSurgePercentage=0. You can read about it more
    [here](https://spot.io/resources/kubernetes-autoscaling/5-kubernetes-deployment-strategies-roll-out-like-the-pros/)
    +label=Rolling update strategy
    """

    type: constr(regex=r"^rolling_update$") = Field(
        ..., description="+value=rolling_update"
    )
    max_unavailable_percentage: conint(ge=0, le=100) = Field(
        25,
        description="+label=Max unavailable(%)\n+usage=Percentage of total replicas that can be brought down at one time.\nFor a value of 25 when replicas are set to 12 this would mean minimum (25% of 12) = 3 pods might be unavailable during the deployment.\nSetting this to a higher value can help in speeding up the deployment process.",
    )
    max_surge_percentage: conint(ge=0, le=100) = Field(
        0,
        description="+label=Max Surge(%)\n+usage=Percentage of total replicas of updated image that can be brought up over the total replicas count.\nFor a value of 25 when replicas are set to 12 this would mean (12+(25% of 12) = 15) pods might be running at one time.\nSetting this to a higher value can help in speeding up the deployment process.",
    )


class ConcurrencyPolicy(str, Enum):
    """
    +usage=Choose whether to allow this job to run while another instance of the job is running, or to replace the currently running instance. Allow
    will enable multiple instances of this job to run. Forbid will keep the current instance of the job running and stop a new instance from being run.
    Replace will terminate any currently running instance of the job and start a new one.
    """

    Forbid = "Forbid"
    Allow = "Allow"
    Replace = "Replace"


class Schedule(BaseModel):
    """
    +docs=Describes that we are going to schedule our job to run at a schedule, making our job a cron job.
    +label=Schedule
    +usage=Run the job on a schedule. [Docs](https://docs.truefoundry.com/docs/deploy-a-cron-job)
    """

    type: constr(regex=r"^scheduled$") = Field(..., description="+value=scheduled")
    schedule: str = Field(
        ...,
        description="+docs=Specify the schedule for this job to be run periodically in cron format. [Learn more](https://docs.truefoundry.com/docs/deploy-a-cron-job)\n+usage=Specify the schedule for this job to be run periodically in cron format.\n```\n* * * * *\n| | | | |\n| | | | |___ day of week (0-6) (Sunday is 0)\n| | | |_____ month (1-12)\n| | |_______ day of month (1-31)\n| |_________ hour (0-23)\n|___________ minute (0-59)\n```",
    )
    concurrency_policy: ConcurrencyPolicy = Field(
        "Forbid",
        description="+usage=Choose whether to allow this job to run while another instance of the job is running, or to replace the currently running instance. Allow\nwill enable multiple instances of this job to run. Forbid will keep the current instance of the job running and stop a new instance from being run.\nReplace will terminate any currently running instance of the job and start a new one.",
    )
    timezone: Optional[str] = Field(
        None,
        description='+usage=Timezone against which the cron schedule will be calculated, e.g. "Asia/Tokyo". Default is machine\'s local time.\nhttps://docs.truefoundry.com/docs/list-of-supported-timezones',
    )


class SecretMount(BaseModel):
    type: constr(regex=r"^secret$") = Field(..., description="+value=secret")
    mount_path: constr(regex=r"^\/(?:[^/\n]+\/*)*[^/\n]+(\.[^/\n]+)?$") = Field(
        ...,
        description="+label=File path\n+usage=Absolute file path where the file will be created.\n+message=Please enter a valid file path",
    )
    secret_fqn: constr(regex=r"^tfy-secret:\/\/.+:.+:.+$") = Field(
        ...,
        description="+label=Secret\n+usage=The Truefoundry secret whose value will be the file content.",
    )


class StringDataMount(BaseModel):
    type: constr(regex=r"^string$") = Field(..., description="+value=string")
    mount_path: constr(regex=r"^\/(?:[^/\n]+\/*)*[^/\n]+(\.[^/\n]+)?$") = Field(
        ...,
        description="+label=File Path\n+usage=Absolute file path where the file will be created.\n+message=Please enter a valid file path",
    )
    data: str = Field(..., description="+label=Data\n+usage=The file content.")


class TruefoundryModelRegistry(BaseModel):
    """
    +docs=Truefoundry model registry (deploy a model from TFY model registry)
    +label=Truefoundry model registry
    +usage=Deploy a model from TFY model registry
    """

    type: constr(regex=r"^tfy-model-registry$") = Field(
        ..., description="+value=tfy-model-registry"
    )
    model_version_fqn: constr(regex=r"^model:[a-zA-Z0-9/_\-]+:\d+$") = Field(
        ...,
        description="+label=Model Version FQN\n+usage=The FQN of the model version that you want to deploy.",
    )


class Autoscaling(BaseModel):
    min_replicas: conint(ge=1) = Field(
        1,
        description="+label=Minimum replicas\n+usage=Minimum number of replicas to keep available",
    )
    max_replicas: conint(ge=1, le=20) = Field(
        ...,
        description="+label=Maximum replicas\n+usage=Maximum number of replicas allowed for the component.",
    )
    metrics: Union[CPUUtilizationMetric, RPSMetric] = Field(
        ...,
        description="+label=Autoscaling metrics\n+usage=Metrics to use for the autoscaler",
    )


class Build(BaseModel):
    """
    +docs=Describes how we build our code into a Docker image.
    +label=Source Code (Build and deploy source code)
    +icon=fa-code
    """

    type: constr(regex=r"^build$") = Field(..., description="+value=build")
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )
    build_source: Union[RemoteSource, GitSource, LocalSource] = Field(
        ...,
        description="+docs=Source code location.\n+label=Fetch source code to build and deploy\n+icon=fa-code\n+sort=1",
    )
    build_spec: Union[DockerFileBuild, PythonBuild] = Field(
        ...,
        description="+docs=Instructions to build a container image out of the build source\n+label=Build using DockerFile or using Buildpack\n+icon=fa-wrench\n+sort=2",
    )


class Canary(BaseModel):
    """
    +docs=This strategy brings up the new release without bringing the older release down. Traffic is shifted from the older release to the newer release in a staged manner.
    This can help with verifying the health of the new release without shifting complete traffic.
    +label=Canary strategy
    """

    type: Literal["canary"] = Field(..., description="+value=canary")
    steps: List[CanaryStep] = Field(
        ...,
        description="+docs=These steps would be executed in order to enable shifting of traffic slowly from stable to canary version\n+label=Steps",
    )


class HealthProbe(BaseModel):
    """
    +docs=Describes the configuration for the Health Probe's
    To learn more you can go [here](https://docs.truefoundry.com/docs/add-health-checks-to-deployments)
    +icon=fa-heart
    """

    config: HttpProbe
    initial_delay_seconds: conint(ge=0, le=36000) = Field(
        0,
        description="+usage=Number of seconds after the container is started before the first probe is initiated.",
    )
    period_seconds: conint(ge=1, le=36000) = Field(
        10, description="+usage=How often, in seconds, to execute the probe."
    )
    timeout_seconds: conint(ge=1, le=36000) = Field(
        1, description="+usage=Number of seconds after which the probe times out."
    )
    success_threshold: conint(ge=1, le=100) = Field(
        1,
        description="+usage=Minimum consecutive successes for the probe to be considered successful after having failed.",
    )
    failure_threshold: conint(ge=1, le=100) = Field(
        3,
        description="+usage=Number of consecutive failures required to determine the container is not alive (liveness probe) or not ready (readiness probe).",
    )


class Job(BaseModel):
    """
    +docs=Describes the configuration for the job
    """

    type: constr(regex=r"^job$") = Field(..., description="+value=job")
    name: constr(regex=r"^[a-z][a-z0-9\-]{1,23}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the job\n+sort=1\n+message=3 to 25 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    image: Union[Build, Image] = Field(
        ...,
        description="+docs=Specify whether you want to deploy a Docker image or build and deploy from source code\n+label=Deploy a Docker image or build and deploy from source code\n+icon=fa-solid fa-cloud-arrow-up:#21B6A8\n+sort=2",
    )
    resources: Optional[Resources] = None
    trigger: Union[Manual, Schedule] = Field(
        {"type": "manual"}, description="+docs=Specify the trigger\n+sort=3"
    )
    retries: conint(ge=0, le=10) = Field(
        0,
        description="+label=Retries\n+usage=Specify the maximum number of attempts to retry a job before it is marked as failed.\n+icon=fa-clone",
    )
    timeout: Optional[conint(le=432000, gt=0)] = Field(
        None,
        description="+label=Timeout\n+usage=Job timeout in seconds.\n+icon=fa-clone",
    )
    params: Optional[List[Param]] = Field(
        None,
        description="+label=Params for input\n+usage=Configure params and pass it to create different job runs\n+icon=fa-globe\n+sort=7",
    )
    env: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Environment Variables\n+usage=Configure environment variables to be injected in the service. [Docs](https://docs.truefoundry.com/docs/env-variables)\n+icon=fa-globe\n+sort=8",
    )
    service_account: Optional[str] = Field(
        None, description="+usage=Service account that this workload should use"
    )
    file_mounts: Optional[List[FileMount]] = Field(
        None,
        description="+usage=Files to be mounted to job pod(s). **Deprecated**\n+ignore\n+sort=10010",
    )
    mounts: Optional[List[Union[SecretMount, StringDataMount]]] = Field(
        None,
        description="+usage=Configure data to be mounted to job pod(s)\n+ignore\n+sort=10011",
    )
    labels: Optional[Dict[str, str]] = Field(
        None, description="+label=Labels\n+usage=Add labels to service metadata"
    )


class ModelDeployment(BaseModel):
    """
    +docs=Describes the configuration for the model deployment
    +usage=To learn more you can go [here](https://dash.readme.com/project/truefoundry/v0.1.1/docs/model-deployment-2)
    """

    type: constr(regex=r"^model-deployment$") = Field(
        ..., description="+value=model-deployment"
    )
    name: constr(regex=r"^[a-z][a-z0-9\-]{1,17}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the model deployment. This uniquely identifies this deployment in the workspace.\n> Name can only contain alphanumeric characters and '-' and can be atmost 20 characters long\n+sort=1\n+label=Model deployment name\n+message=3 to 19 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    model_source: Union[TruefoundryModelRegistry, HuggingfaceModelHub] = Field(
        ...,
        description="+docs=Specify the model to be deployed\n+sort=2\n+label=Model to deploy",
    )
    resources: Optional[Resources] = None
    grpc: bool = Field(False, description="+docs=Use grpc\n+sort=4\n+label=Use gRPC")
    labels: Optional[Dict[str, str]] = Field(
        None, description="+label=Labels\n+usage=Add labels to service metadata"
    )


class Service(BaseModel):
    """
    +docs=Describes the configuration for the service
    """

    type: constr(regex=r"^service$") = Field(..., description="+value=service")
    name: constr(regex=r"^[a-z][a-z0-9\-]{1,23}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the service. This uniquely identifies this service in the workspace.\n> Name can only contain alphanumeric characters and '-' and can be atmost 25 characters long\n+sort=1\n+message=3 to 25 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    image: Union[Build, Image] = Field(
        ...,
        description="+docs=Specify whether you want to deploy a Docker image or build and deploy from source code\n+label=Deploy a Docker image or build and deploy from source code\n+icon=fa-solid fa-cloud-arrow-up:#21B6A8\n+sort=2",
    )
    resources: Optional[Resources] = None
    replicas: Union[confloat(ge=1.0, le=20.0), Autoscaling] = Field(
        1,
        description="+label=Replicas\n+usage=Replicas of service you want to run\n+icon=fa-clone\n+sort=3",
    )
    env: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Environment Variables\n+usage=Configure environment variables to be injected in the service. [Docs](https://docs.truefoundry.com/docs/env-variables)\n+icon=fa-globe\n+sort=6",
    )
    ports: List[Port] = Field(
        ...,
        description="+docs=Specify the ports you want the service to be exposed to\n+label=Ports (Ports to route customer traffic to)\n+icon=fa-plug\n+sort=4",
    )
    liveness_probe: Optional[HealthProbe] = None
    readiness_probe: Optional[HealthProbe] = None
    service_account: Optional[str] = Field(
        None, description="+usage=Service account that this workload should use"
    )
    file_mounts: Optional[List[FileMount]] = Field(
        None,
        description="+usage=Files to be mounted to service pod(s). **Deprecated**\n+ignore\n+sort=10010",
    )
    mounts: Optional[List[Union[SecretMount, StringDataMount]]] = Field(
        None,
        description="+usage=Configure data to be mounted to service pod(s)\n+ignore\n+sort=10011",
    )
    labels: Optional[Dict[str, str]] = Field(
        None, description="+label=Labels\n+usage=Add labels to service metadata"
    )
    rollout_strategy: Optional[Union[Rolling, Canary, BlueGreen]] = Field(
        None,
        description="+label=Rollout strategy\n+usage=Strategy to dictate how a rollout should happen when a new release for this service is made",
    )


class Application(BaseModel):
    __root__: Union[Service, Job, ModelDeployment]



import json
import os

import questionary

from servicefoundry.lib.clients.terragrunt_client import Terragrunt
from servicefoundry.lib.config.config_manager import ConfigManager
from servicefoundry.lib.config.dict_questionaire import DictQuestionaire


class InfraConfig:
    def __init__(
        self, terragrunt_client: Terragrunt, infra_config_file_path=None
    ) -> None:
        self.terragrunt_client = terragrunt_client
        if infra_config_file_path and os.path.exists(infra_config_file_path):
            with open(infra_config_file_path, "r") as config:
                infra_config = json.loads(config.read())
        else:
            infra_config = ConfigManager().get_config(ConfigManager.INFRA_CONFIG)
        if infra_config:
            self.provisioning = infra_config["provisioning"]
            self.bootstrapping = infra_config["bootstrapping"]

    def __external_config_handler(self):
        return questionary.confirm(
            f"Do you want to attach your own network: ", default=False
        ).ask()

    # Config needed for the provisioning step where terraform is run
    provisioning = {
        "provider": None,
        "awsInputs": {
            "remoteStateBucket": None,
            "region": "eu-west-1",
            "shortRegion": "euwe1",
            "remoteStateAwsProfile": None,
            "awsProfile": None,
            "accountId": None,
            "accountName": None,
            "clusterPrefix": "tfy-",
            "baseDomain": "truefoundry.com",
            "externalNetworkConfig": {
                "networkVpcId": None,
                "networkPublicSubnetsId": [],
                "networkPrivateSubnetsId": [],
            },
        },
    }
    # Config needed for the boostrapping step where helm apply is run
    bootstrapping = {
        "cluster": {
            "provider": None,
            "name": None,
            "region": None,
            "shortRegion": None,
            "baseDomain": None,
            "account": None,
            "endpoint": None,
        },
        "argoIamRole": None,
        "certificatesIamRole": None,
        "externalDnsIamRole": None,
        "awsComponents": {
            "awsEbsCsiDriverIamRole": None,
            "karpenter": {"iamRole": None, "instanceProfile": None},
        },
        "workloadComponents": {
            "tfyAgent": {
                "tenantName": None,
                "clusterToken": None,
                "controlPlaneUrl": None,
            }
        },
        "targetRepo": {
            "url": None,
            "branch": None,
            "path": None,
            "username": None,
            "password": None,
        },
        "monitoring": {
            "loki": {"bucketName": None, "roleArn": None},
            "prometheus": {"bucketName": None, "roleArn": None},
        },
        "controlPlaneComponents": {
            "controlPlaneUrl": None,
            "mlfoundry": {
                "enabled": False,
                "s3BucketName": None,
            },
            "mlmonitoring": {
                "enabled": False,
            },
            "servicefoundry": {
                "enabled": True,
                "s3BucketName": None,
            },
        },
    }

    def populate_bootstrapping_config(self, target_repo_config, base_terragrunt_dir):
        self.bootstrapping["cluster"]["account"] = self.provisioning["awsInputs"][
            "accountName"
        ]
        self.bootstrapping["cluster"]["provider"] = self.provisioning["provider"]
        self.bootstrapping["cluster"][
            "name"
        ] = self.terragrunt_client.fetch_terragrunt_output(
            os.path.join(base_terragrunt_dir, "cluster"), "cluster_id"
        )
        self.bootstrapping["cluster"]["region"] = self.provisioning["awsInputs"][
            "region"
        ]
        self.bootstrapping["cluster"]["shortRegion"] = self.provisioning["awsInputs"][
            "shortRegion"
        ]
        self.bootstrapping["cluster"]["baseDomain"] = self.provisioning["awsInputs"][
            "baseDomain"
        ]
        self.bootstrapping["cluster"][
            "endpoint"
        ] = self.terragrunt_client.fetch_terragrunt_output(
            os.path.join(base_terragrunt_dir, "cluster"), "cluster_endpoint"
        )

        self.bootstrapping[
            "argoIamRole"
        ] = self.terragrunt_client.fetch_terragrunt_output(
            os.path.join(base_terragrunt_dir, "cluster-app-argocd"),
            "argocd_iam_role_arn",
        )
        if not self.provisioning["awsInputs"]["externalNetworkConfig"]:
            self.bootstrapping[
                "certificatesIamRole"
            ] = self.terragrunt_client.fetch_terragrunt_output(
                os.path.join(base_terragrunt_dir, "cluster-iam-certmanager"),
                "iam_role_arn",
            )
            self.bootstrapping[
                "externalDnsIamRole"
            ] = self.terragrunt_client.fetch_terragrunt_output(
                os.path.join(base_terragrunt_dir, "cluster-iam-external-dns"),
                "iam_role_arn",
            )
        else:
            self.bootstrapping["certificatesIamRole"] = ""
            self.bootstrapping["externalDnsIamRole"] = ""
        self.bootstrapping["awsComponents"][
            "awsEbsCsiDriverIamRole"
        ] = self.terragrunt_client.fetch_terragrunt_output(
            os.path.join(base_terragrunt_dir, "cluster-iam-csi-ebs"), "iam_role_arn"
        )
        self.bootstrapping["awsComponents"]["karpenter"][
            "iamRole"
        ] = self.terragrunt_client.fetch_terragrunt_output(
            os.path.join(base_terragrunt_dir, "cluster-iam-karpenter"), "iam_role_arn"
        )
        self.bootstrapping["awsComponents"]["karpenter"][
            "instanceProfile"
        ] = self.terragrunt_client.fetch_terragrunt_output(
            os.path.join(base_terragrunt_dir, "cluster-iam-karpenter"),
            "iam_instance_profile_id",
        )
        install_workload = questionary.confirm(
            "Do you want to install workload components as well?", default=False
        ).ask()
        if install_workload:
            self.bootstrapping["workloadComponents"]["tfyAgent"][
                "tenantName"
            ] = questionary.text("Tenant name").ask()
            self.bootstrapping["workloadComponents"]["tfyAgent"][
                "clusterToken"
            ] = questionary.text("Cluster token").ask()
            self.bootstrapping["workloadComponents"]["tfyAgent"][
                "controlPlaneUrl"
            ] = questionary.text("Control plane url").ask()
        else:
            self.bootstrapping["workloadComponents"] = None

        install_control_plane = questionary.confirm(
            "Do you want to install control plane components as well?", default=False
        ).ask()
        if install_control_plane:
            self.bootstrapping["controlPlaneComponents"][
                "controlPlaneUrl"
            ] = questionary.text("Control plane url").ask()
            self.bootstrapping["controlPlaneComponents"]["mlfoundry"][
                "enabled"
            ] = questionary.confirm(
                "Do you want to install mlfoundry", default=False
            ).ask()
            if self.bootstrapping["controlPlaneComponents"]["mlfoundry"]["enabled"]:
                self.bootstrapping["controlPlaneComponents"]["mlfoundry"][
                    "s3BucketName"
                ] = self.terragrunt_client.fetch_terragrunt_output(
                    os.path.join(base_terragrunt_dir, "truefoundry-aws"),
                    "mlfoundry_bucket_id",
                )
            self.bootstrapping["controlPlaneComponents"]["mlmonitoring"][
                "enabled"
            ] = questionary.confirm(
                "Do you want to install mlmonitoring", default=False
            ).ask()
            self.bootstrapping["controlPlaneComponents"]["servicefoundry"][
                "enabled"
            ] = questionary.confirm(
                "Do you want to install servicefoundry", default=False
            ).ask()
            if self.bootstrapping["controlPlaneComponents"]["servicefoundry"][
                "enabled"
            ]:
                self.bootstrapping["controlPlaneComponents"]["servicefoundry"][
                    "s3BucketName"
                ] = self.terragrunt_client.fetch_terragrunt_output(
                    os.path.join(base_terragrunt_dir, "truefoundry-aws"),
                    "svcfoundry_bucket_id",
                )
            else:
                self.bootstrapping["controlPlaneComponents"]["servicefoundry"] = None
        else:
            self.bootstrapping["controlPlaneComponents"] = None

        self.bootstrapping["targetRepo"] = target_repo_config

        install_monitoring = questionary.confirm(
            "Do you want to install monitoring as well?", default=False
        ).ask()
        if install_monitoring:
            self.bootstrapping["monitoring"]["loki"][
                "bucketName"
            ] = self.terragrunt_client.fetch_terragrunt_output(
                os.path.join(base_terragrunt_dir, "cluster-app-loki"),
                "loki_bucket_name",
            )
            self.bootstrapping["monitoring"]["loki"][
                "roleArn"
            ] = self.terragrunt_client.fetch_terragrunt_output(
                os.path.join(base_terragrunt_dir, "cluster-app-loki"), "iam_role_arn"
            )
            self.bootstrapping["monitoring"]["prometheus"][
                "bucketName"
            ] = self.terragrunt_client.fetch_terragrunt_output(
                os.path.join(base_terragrunt_dir, "cluster-app-thanos"),
                "thanos_bucket_name",
            )
            self.bootstrapping["monitoring"]["prometheus"][
                "roleArn"
            ] = self.terragrunt_client.fetch_terragrunt_output(
                os.path.join(base_terragrunt_dir, "cluster-app-thanos"), "iam_role_arn"
            )
        else:
            self.bootstrapping["monitoring"] = None
        ConfigManager().save_config(ConfigManager.INFRA_CONFIG, self.to_json())

    def populate_provisioning_config(self, provider: str):
        self.provisioning["provider"] = "aws"
        DictQuestionaire(self.provisioning["awsInputs"]).ask(
            custom_handlers={"externalNetworkConfig": self.__external_config_handler}
        )
        ConfigManager().save_config(ConfigManager.INFRA_CONFIG, self.to_json())

    def to_json(self):
        return {"provisioning": self.provisioning, "bootstrapping": self.bootstrapping}



import json

import questionary


class DictQuestionaire:
    def __init__(self, obj: dict) -> None:
        self.obj = obj

    def __confirm_val(self, v, tree_path, custom_handlers=None):
        custom_handlers = custom_handlers or {}
        if tree_path in custom_handlers:
            if not custom_handlers[tree_path]():
                return

        if v is None:
            v = questionary.text(f"Please Enter {tree_path} : ").ask()
        else:
            v = questionary.text(f"Please Enter {tree_path} : ", default=str(v)).ask()
        return v

    def __confirm_list_val(self, v, tree_path, custom_handlers=None):
        custom_handlers = custom_handlers or {}
        if tree_path in custom_handlers:
            if not custom_handlers[tree_path]():
                return

        temp_v = questionary.text(f"Please Enter {tree_path} : ", default=str(v)).ask()
        temp_v = str(temp_v).replace("'", '"')
        v = json.loads(temp_v)
        return v

    # TODO: add unit test for get_all_values function
    def __get_all_values(self, dicts, tree_path="", custom_handlers=None):
        custom_handlers = custom_handlers or {}
        if tree_path in custom_handlers:
            if not custom_handlers[tree_path]():
                return None
        if tree_path != "":
            tree_path += "."

        for k, v in dicts.items():
            if isinstance(v, dict):
                dicts[k] = self.__get_all_values(v, tree_path + k, custom_handlers)
            elif isinstance(v, list):
                if len(v) != 0 and isinstance(v[0], dict):
                    for i, vi in enumerate(v):
                        self.__get_all_values(
                            vi, tree_path + k + "[" + i + "]", custom_handlers
                        )
                else:
                    dicts[k] = self.__confirm_list_val(
                        v, tree_path + k, custom_handlers
                    )
            else:
                dicts[k] = self.__confirm_val(v, tree_path + k, custom_handlers)
        return dicts

    def ask(self, custom_handlers=None):
        custom_handlers = custom_handlers or {}
        return self.__get_all_values(self.obj, custom_handlers=custom_handlers)



import sys

from servicefoundry.cli import create_servicefoundry_cli


def main():
    # Exit the interpreter by raising SystemExit(status).
    # If the status is omitted or None, it defaults to zero (i.e., success).
    # If the status is an integer, it will be used as the system exit status.
    # If it is another kind of object, it will be printed and the system exit status will be one (i.e., failure).
    sys.exit(create_servicefoundry_cli()())


if __name__ == "__main__":
    main()



import json
import os


class ConfigManager:
    INFRA_CONFIG = "infra_config"
    TARGET_REPO_CONFIG = "target_repo_config"
    INSTALLATION_CONFIG = "installation_config"

    __config_dir = os.path.join(os.path.expanduser("~"), ".truefoundry", "config")

    __config_file_map = {
        INFRA_CONFIG: "infra-config.json",
        INSTALLATION_CONFIG: "installation-config.json",
        TARGET_REPO_CONFIG: "target-repo-config.json",
    }

    def __init__(self) -> None:
        os.makedirs(self.__config_dir, exist_ok=True)

    def get_config(self, config_key: str) -> any:
        if config_key not in self.__config_file_map:
            raise Exception(f"Config key {config_key} not found")

        file_path = os.path.join(self.__config_dir, self.__config_file_map[config_key])
        if os.path.exists(file_path):
            with open(file_path, "r") as config:
                return json.loads(config.read())
        else:
            return None

    def save_config(self, config_key: str, config: any):
        if config_key not in self.__config_file_map:
            raise Exception(f"Config key {config_key} not found")

        file_path = os.path.join(self.__config_dir, self.__config_file_map[config_key])
        with open(file_path, "w") as config_file:
            config_file.write(json.dumps(config))



class BadRequestException(Exception):
    def __init__(self, status_code, message=None):
        super(BadRequestException, self).__init__()
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return self.message


class ConfigurationException(Exception):
    def __init__(self, message=None):
        super(ConfigurationException, self).__init__()
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return self.message



SFY = "sfy"

# TODO: probably create another `rich_messages.py` and apply all formatting there
PROMPT_LOGIN_SUCCESSFUL = f"""[green bold]Login Successful![/]"""
PROMPT_LOGIN_INFO = (
    f"""[yellow]You are logged in as {{username!r}} with email {{email!r}}[/]"""
)
PROMPT_LOGOUT_SUCCESSFUL = f"""[green bold]Logged Out![/]"""
PROMPT_POST_LOGIN = ""
PROMPT_ALREADY_LOGGED_OUT = f"""[yellow]You are already logged out[/]"""
PROMPT_USING_CLUSTER_CONTEXT = f"""[yellow]Using cluster {{!r}}[/]"""
PROMPT_USING_WORKSPACE_CONTEXT = f"""[yellow]Using workspace {{!r}}[/]"""
PROMPT_USING_APPLICATION_CONTEXT = f"""[yellow]Using application {{!r}}[/]"""
PROMPT_CREATING_NEW_WORKSPACE = f"""[yellow]Creating a new workspace {{!r}}[/]"""
PROMPT_DELETING_WORKSPACE = f"""[yellow]Deleting workspace {{!r}}[/]"""
PROMPT_DELETED_WORKSPACE = f"""[green]Deleted workspace {{!r}}[/]"""
PROMPT_DELETING_SERVICE = f"""[yellow]Deleting service {{!r}}[/]"""
PROMPT_DELETED_SERVICE = f"""[green]Deleted service {{!r}}[/]"""
PROMPT_NO_WORKSPACES = f"""[yellow]No workspaces found. Either cluster name is wrong, or your cluster doesn't contain any workspaces. You can create one with [bold]{SFY} create workspace[/][/]"""
PROMPT_NO_APPLICATIONS = f"""[yellow]No applications found. You can create one with [bold]{SFY} deploy[/] from within your application folder"""
PROMPT_NO_VERSIONS = f"""[yellow]No application versions found."""



import subprocess
from typing import Optional

import rich_click as click

from servicefoundry.cli.const import GROUP_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.io.rich_output_callback import RichOutputCallBack


@click.group(
    name="patch",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Patch parts of servicefoundry.yaml file",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    default="./servicefoundry.yaml",
    help="Path to servicefoundry.yaml file",
    show_default=True,
)
@click.option(
    "--filter",
    type=click.STRING,
    required=True,
    help="`yq` filter",
)
@click.option(
    "-o",
    "--output-file",
    type=click.Path(exists=False, writable=True, resolve_path=True),
    help="Path to yaml file to redirect the output",
    show_default=True,
)
@click.option(
    "-I",
    "--indent",
    default=4,
    type=click.INT,
    help="Indent level for output",
    show_default=True,
)
@handle_exception_wrapper
def patch_command(
    file: str, filter: str, output_file: Optional[str], indent: int
) -> None:
    yq_command = [
        "yq",
        "--output-format",
        "yaml",
        "--indent",
        str(indent),
        filter,
        file,
    ]
    if output_file:
        with open(output_file, "w") as fd:
            subprocess.run(yq_command, stdout=fd)
    else:
        p = subprocess.run(yq_command, stdout=subprocess.PIPE)
        output = p.stdout.decode("UTF-8")
        callback = RichOutputCallBack()
        callback.print_line(output)


def get_patch_command():
    return patch_command



import os
import tarfile
import tempfile
import time
import warnings
from typing import Callable, List, Optional

import gitignorefile
from tqdm import tqdm

from servicefoundry import builder
from servicefoundry.auto_gen import models
from servicefoundry.builder.docker_service import (
    pull_docker_image,
    push_docker_image,
    push_docker_image_with_latest_tag,
)
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.logger import logger
from servicefoundry.v2.lib.patched_models import Image, RemoteSource


def _human_readable_size(num_bytes: float) -> str:
    units = ["B", "KiB", "MiB", "GiB"]
    i = 0
    amount = num_bytes
    while amount >= 1024 and i < len(units) - 1:
        amount /= 1024.0
        i += 1
    amount = round(amount, 2)
    return f"{amount} {units[i]}"


def _make_tarfile(
    output_filename: str,
    source_dir: str,
    additional_directories: List[str],
    is_path_ignored: Optional[Callable[[str], bool]] = None,
) -> None:
    if not is_path_ignored:
        # if no callback handler present assume that every file needs to be added
        is_path_ignored = lambda *_: False

    with tarfile.open(output_filename, "w:gz") as tar:
        source_dirs = [source_dir]
        source_dirs.extend(additional_directories)
        for source_dir in source_dirs:
            _add_files_in_tar(
                is_path_ignored=is_path_ignored,
                source_dir=source_dir,
                tar=tar,
            )


def _add_files_in_tar(
    is_path_ignored: Callable[[str], bool],
    source_dir: str,
    tar: tarfile.TarFile,
) -> None:
    for root, dirs, files in tqdm(
        os.walk(source_dir, topdown=True), desc="Packaging source code"
    ):
        if is_path_ignored(root):
            logger.debug("Ignoring directory %s", root)

            # NOTE: we can safely ignore going through the sub-dir
            # if root itself is excluded.
            dirs.clear()
            continue
        logger.debug("Adding contents of the directory %s", root)
        files_added_count = 0
        for file in files:
            file_path = os.path.join(root, file)
            if not is_path_ignored(file_path):
                arcname = os.path.relpath(file_path, source_dir)
                tar.add(file_path, arcname=arcname)
                logger.debug("Adding %s with arcname %r", file_path, arcname)
                files_added_count += 1

        if not files_added_count:
            # If no files were added and the directory was not ignored too, then we
            # want to add the directory as an empty dir
            arcname = os.path.relpath(root, source_dir)
            tar.add(root, arcname=arcname, recursive=False)
            logger.debug("Adding empty directory %s with arcname %r", root, arcname)


def _get_callback_handler_to_ignore_file_path(
    source_dir: str,
) -> Optional[Callable[[str], bool]]:
    ignorefile_path = os.path.join(source_dir, ".tfyignore")
    if os.path.exists(ignorefile_path):
        logger.info(".tfyignore file found in %s", source_dir)
        return gitignorefile.parse(path=ignorefile_path, base_path=source_dir)

    ignorefile_path = os.path.join(source_dir, ".sfyignore")
    if os.path.exists(ignorefile_path):
        logger.info(".sfyignore file found in %s", source_dir)
        warnings.warn(
            "`.sfyignore` is deprecated and will be ignored in future versions. "
            "Please rename the file to `.tfyignore`",
            category=DeprecationWarning,
        )
        return gitignorefile.parse(path=ignorefile_path, base_path=source_dir)

    # check for valid git repo
    try:
        import git

        repo = git.Repo(source_dir, search_parent_directories=True)
        logger.info(
            "Git repository detected in source. Files added in .gitignore will be ignored.\n"
            "If you don't want to ignore these files or otherwise, "
            "please create .tfyignore file and add file patterns to ignore."
        )
        return lambda file_path: bool(repo.ignored(file_path))
    except Exception as ex:
        logger.debug(
            "Could not treat source %r as a git repository due to %r", source_dir, ex
        )

    logger.info(
        "Neither `.tfyignore` file found in %s nor a valid git repository found. "
        "We recommend you to create .tfyignore file and add file patterns to ignore",
        source_dir,
    )
    return None


def local_source_to_remote_source(
    local_source: models.LocalSource,
    workspace_fqn: str,
    component_name: str,
) -> RemoteSource:
    with tempfile.TemporaryDirectory() as local_dir:
        package_local_path = os.path.join(local_dir, "build.tar.gz")
        source_dir = os.path.abspath(local_source.project_root_path)

        if not os.path.exists(source_dir):
            raise ValueError(
                f"project root path {source_dir!r} of component {component_name!r} does not exist"
            )

        logger.info("Archiving contents of dir: %r", source_dir)

        is_path_ignored = _get_callback_handler_to_ignore_file_path(source_dir)
        _make_tarfile(
            output_filename=package_local_path,
            source_dir=source_dir,
            additional_directories=[],
            is_path_ignored=is_path_ignored,
        )

        try:
            file_size = _human_readable_size(os.path.getsize(package_local_path))
            logger.info("Code archive size: %r", file_size)
        except Exception:
            # Should not block code upload
            logger.exception("Failed to calculate code archive size")

        logger.debug("Uploading code archive.")
        client = ServiceFoundryServiceClient()
        remote_uri = client.upload_code_package(
            workspace_fqn=workspace_fqn,
            component_name=component_name,
            package_local_path=package_local_path,
        )
        logger.debug("Uploaded code archive.")
        return RemoteSource(remote_uri=remote_uri)


def local_source_to_image(
    build: models.Build,
    docker_registry_fqn: Optional[str],
    workspace_fqn: str,
    component_name: str,
) -> Image:
    source_dir = os.path.abspath(
        os.path.expanduser(build.build_source.project_root_path)
    )
    if not os.path.exists(source_dir):
        raise ValueError(
            f"project root path {source_dir!r} of component {component_name!r} does not exist"
        )

    build.build_spec.build_context_path = os.path.join(
        source_dir, build.build_spec.build_context_path
    )
    if not os.path.exists(build.build_spec.build_context_path):
        raise ValueError(
            f"Build context path {build.build_spec.build_context_path!r} "
            f"of component {component_name!r} does not exist"
        )

    if isinstance(build.build_spec, models.DockerFileBuild):
        build.build_spec.dockerfile_path = os.path.join(
            source_dir, build.build_spec.dockerfile_path
        )
        if not os.path.exists(build.build_spec.dockerfile_path):
            raise ValueError(
                f"Dockerfile path {build.build_spec.dockerfile_path!r} "
                f"of component {component_name!r} does not exist"
            )

    client = ServiceFoundryServiceClient()
    docker_registry = client.get_docker_registry_creds(
        docker_registry_fqn=docker_registry_fqn,
    )

    docker_registry_fqn = docker_registry.fqn
    registry_url = docker_registry.registryUrl
    registry_username = docker_registry.username
    registry_password = docker_registry.password

    create_repo_response = client.create_repo_in_registry(
        docker_registry_fqn=docker_registry_fqn,
        workspace_fqn=workspace_fqn,
        application_name=component_name,
    )
    repo_name = create_repo_response.repoName

    image_uri = f"{registry_url}/{repo_name}:{int(time.time())}"
    cache_from = f"{registry_url}/{repo_name}:latest"

    # pull cache image if it exists
    try:
        pull_docker_image(
            image_uri=cache_from,
            docker_login_username=registry_username,
            docker_login_password=registry_password,
        )
    except Exception:
        logger.info(f"Failed to pull {cache_from}. Building without cache image.")

    # build
    builder.build(
        build.build_spec,
        tag=image_uri,
        cache_from=cache_from,
    )

    # push built image to registry
    push_docker_image(
        image_uri=image_uri,
        docker_login_username=registry_username,
        docker_login_password=registry_password,
    )

    # push 'latest' tag for image
    push_docker_image_with_latest_tag(
        image_uri=image_uri,
        docker_login_username=registry_username,
        docker_login_password=registry_password,
    )

    return Image(image_uri=image_uri, docker_registry=docker_registry_fqn)



import re
from datetime import datetime, timedelta, timezone

from dateutil import parser
from dateutil.tz import tzlocal

from servicefoundry.logger import logger

time_duration_regex = re.compile("^[0-9]+[smhd]$")

granularity_timedelta = {
    "s": timedelta(seconds=1),
    "m": timedelta(minutes=1),
    "h": timedelta(hours=1),
    "d": timedelta(days=1),
}


def get_timestamp_from_timestamp_or_duration(time_or_duration: str):
    """
    Returns timestamp in milliseconds
    """
    if time_duration_regex.match(time_or_duration):
        number = int(time_or_duration[:-1])
        granularity = time_or_duration[-1]
        current_time = datetime.now(timezone.utc)
        required_time = current_time - granularity_timedelta[granularity] * number
        return required_time.timestamp() * 1000
    else:
        logger.debug(
            f"Cannot parse: {time_or_duration} as duration, regex match failed"
        )
        try:
            date_time_obj = parser.parse(time_or_duration)
        except ValueError as e:
            logger.debug(f"Value not a valid timestamp or duration: {time_or_duration}")
            raise ValueError(
                f"Value not a valid timestamp or duration: {time_or_duration}"
            )
        if not date_time_obj.tzinfo:
            date_time_obj = date_time_obj.replace(tzinfo=tzlocal())
        date_time_obj = date_time_obj.astimezone(timezone.utc)
        return date_time_obj.timestamp() * 1000



from abc import ABC, abstractmethod

from servicefoundry.io.parameters import OptionsParameter, Parameter


class InputHook(ABC):
    @abstractmethod
    def confirm(self, prompt, default=False):
        pass

    @abstractmethod
    def ask_string(self, param: Parameter):
        pass

    @abstractmethod
    def ask_number(self, param: Parameter):
        pass

    def ask_integer(self, param: Parameter):
        return self.ask_number(param)

    def ask_float(self, param: Parameter):
        return self.ask_number(param)

    @abstractmethod
    def ask_file_path(self, param: Parameter):
        pass

    @abstractmethod
    def ask_option(self, param: OptionsParameter):
        pass

    @abstractmethod
    def ask_python_file(self, param: Parameter):
        pass


class FailInputHook(InputHook):
    def confirm(self, prompt, default=False):
        raise NotImplementedError(
            f"{InputHook.__name__} doesn't implement method ask_boolean"
        )

    def ask_string(self, param: Parameter):
        raise NotImplementedError(
            f"{InputHook.__name__} doesn't implement method ask_string"
        )

    def ask_number(self, param: Parameter):
        raise NotImplementedError(
            f"{InputHook.__name__} doesn't implement method ask_number"
        )

    def ask_file_path(self, param: Parameter):
        raise NotImplementedError(
            f"{InputHook.__name__} doesn't implement method ask_number"
        )

    def ask_option(self, param: OptionsParameter):
        raise NotImplementedError(
            f"{InputHook.__name__} doesn't implement method ask_option"
        )

    def ask_python_file(self, param: Parameter):
        raise NotImplementedError(
            f"{InputHook.__name__} doesn't implement method ask_file"
        )



import os

from cookiecutter.main import cookiecutter


class CookieCutter:
    def __init__(self, template_path: str, directory: str = ""):
        self.template_path = template_path
        self.directory = directory

    def run(self, destination_dir):
        current_dir = os.getcwd()
        os.chdir(os.path.join(destination_dir))
        if self.directory != "":
            generated_repo_path = cookiecutter(
                self.template_path, no_input=True, directory=self.directory
            )
        else:
            generated_repo_path = cookiecutter(self.template_path, no_input=True)

        os.chdir(current_dir)
        return generated_repo_path



import logging
from typing import Optional

import rich_click as click

from servicefoundry.cli.const import COMMAND_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.io.rich_output_callback import RichOutputCallBack
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.util import get_deployment_fqn_from_application_fqn
from servicefoundry.logger import logger


@click.command(name="build-logs", cls=COMMAND_CLS)
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    help="Application FQN",
    required=True,
)
@click.option(
    "--version",
    type=click.INT,
    help="Deployment version. If no deployment version "
    "given, logs will be fetched for the latest version",
)
@click.option("-f", "--follow", help="Follow log output", is_flag=True, default=False)
@handle_exception_wrapper
def build_logs_command(
    application_fqn: str,
    version: Optional[int],
    follow: bool,
) -> None:
    """
    Get build logs for application fqn and deployment version
    """
    output_hook = RichOutputCallBack()
    tfs_client = ServiceFoundryServiceClient()

    if not version:
        app_fqn_response = tfs_client.get_application_info_by_fqn(application_fqn)
        application_info = tfs_client.get_application_info(
            app_fqn_response.applicationId
        )
        version = application_info.lastVersion

    deployment_fqn = get_deployment_fqn_from_application_fqn(
        application_fqn=application_fqn, version=version
    )

    deployment_fqn_response = tfs_client.get_deployment_info_by_fqn(
        deployment_fqn=deployment_fqn
    )

    build_responses = tfs_client.get_deployment_build_response(
        application_id=deployment_fqn_response.applicationId,
        deployment_id=deployment_fqn_response.deploymentId,
    )

    if not build_responses:
        raise Exception(
            f"Unable to find a build version for application fqn {application_fqn} and version {version}"
        )

    build_response = build_responses[0]
    if not follow:
        tfs_client.fetch_build_logs(
            build_response=build_response,
            callback=output_hook,
        )
    else:
        try:
            logger.info(
                "You can press Ctrl + C to exit the tailing of build "
                "logs and deployment will continue on the server"
            )
            tfs_client.tail_build_logs(
                build_response=build_response,
                callback=output_hook,
                wait=True,
            )
        except KeyboardInterrupt:
            logger.info("Ctrl-C executed. The deployment will still continue.")


def get_build_logs_command():
    return build_logs_command



import os

import rich_click as click

DISPLAY_DATETIME_FORMAT = "%d %b %Y %H:%M:%S %Z"

# TODO (chiragjn): This is a hacky solution used while generating docs. Although
#                  this does not cover cases where custom cls class already inherited from rich_click classes
#                  is being used
RICH_CLICK_DISABLED = os.getenv("RICH_CLICK_DISABLED")
GROUP_CLS = click.Group if RICH_CLICK_DISABLED else click.RichGroup
COMMAND_CLS = click.Command if RICH_CLICK_DISABLED else click.RichCommand



from typing import List

import numpy as np


def normal(loc: float, scale: float, size: List[int]):
    return np.random.normal(loc=loc, scale=scale, size=size).tolist()


def uniform(low: float, high: float, size: List[int]):
    return np.random.uniform(low=low, high=high, size=size).tolist()



import logging

import rich_click as click

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_entity_obj, print_json
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import application as application_lib
from servicefoundry.lib.dao import version as version_lib
from servicefoundry.lib.dao import workspace as workspace_lib
from servicefoundry.logger import logger

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="get", cls=GROUP_CLS)
def get_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Get Truefoundry resources

    \b
    Supported resources:
    - Workspace
    - Application
    - Application Version
    """
    pass


@click.command(name="workspace", cls=COMMAND_CLS, help="Get Workspace details")
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the Workspace",
    required=True,
)
@handle_exception_wrapper
def get_workspace(workspace_fqn):
    workspace = workspace_lib.get_workspace_by_fqn(workspace_fqn=workspace_fqn)
    if CliConfig.get("json"):
        print_json(data=workspace.dict())
    else:
        print_entity_obj("Workspace", workspace)


@click.command(name="application", cls=COMMAND_CLS, help="Get Application details")
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the application",
    required=True,
)
@handle_exception_wrapper
def get_application(application_fqn):
    application = application_lib.get_application(application_fqn=application_fqn)
    if CliConfig.get("json"):
        print_json(data=application.dict())
    else:
        print_entity_obj(
            "Application",
            application,
        )


@click.command(
    name="application-version", cls=COMMAND_CLS, help="Get Application Version details"
)
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the application",
    required=True,
)
@click.option(
    "--version",
    type=click.STRING,
    default=None,
    help="Version number of the application deployment",
    required=True,
)
@handle_exception_wrapper
def get_version(application_fqn, version):
    version = version_lib.get_version(application_fqn=application_fqn, version=version)
    if CliConfig.get("json"):
        print_json(data=version.dict())
    else:
        print_entity_obj("Version", version)


def get_get_command():
    get_command.add_command(get_workspace)
    get_command.add_command(get_application)
    get_command.add_command(get_version)
    return get_command



import os
import threading
from abc import ABC, abstractmethod

from servicefoundry.lib.auth.auth_service_client import AuthServiceClient
from servicefoundry.lib.auth.credential_file_manager import (
    CredentialsFileContent,
    CredentialsFileManager,
)
from servicefoundry.lib.const import API_KEY_ENV_NAME
from servicefoundry.lib.model.entity import Token
from servicefoundry.logger import logger

TOKEN_REFRESH_LOCK = threading.RLock()

from servicefoundry.lib.clients.utils import resolve_base_url


class CredentialProvider(ABC):
    @property
    @abstractmethod
    def token(self) -> Token:
        ...

    @property
    @abstractmethod
    def base_url(self) -> str:
        ...

    @staticmethod
    @abstractmethod
    def can_provide() -> bool:
        ...


class EnvCredentialProvider(CredentialProvider):
    def __init__(self) -> None:
        from servicefoundry.lib.clients.service_foundry_client import (
            ServiceFoundryServiceClient,
        )

        logger.debug("Using env var credential provider")
        api_key = os.getenv(API_KEY_ENV_NAME)
        if not api_key:
            raise Exception(
                f"Value of {API_KEY_ENV_NAME} env var should be non-empty string"
            )
        # TODO: Read host from cred file as well.
        base_url = resolve_base_url().strip("/")
        self._host = base_url
        self._auth_service = AuthServiceClient(base_url=base_url)

        service_foundry_client = ServiceFoundryServiceClient(
            init_session=False, base_url=base_url
        )
        self._token: Token = service_foundry_client.get_token_from_api_key(
            api_key=api_key
        )

    @staticmethod
    def can_provide() -> bool:
        return API_KEY_ENV_NAME in os.environ

    @property
    def token(self) -> Token:
        with TOKEN_REFRESH_LOCK:
            if self._token.is_going_to_be_expired():
                logger.info("Refreshing access token")
                self._token = self._auth_service.refresh_token(self._token)
            return self._token

    @property
    def base_url(self) -> str:
        return self._host


class FileCredentialProvider(CredentialProvider):
    def __init__(self) -> None:
        logger.debug("Using file credential provider")
        self._cred_file = CredentialsFileManager()

        with self._cred_file:
            self._last_cred_file_content = self._cred_file.read()
            self._token = self._last_cred_file_content.to_token()
            self._host = self._last_cred_file_content.host

        self._auth_service = AuthServiceClient(base_url=self._host)

    @staticmethod
    def can_provide() -> bool:
        with CredentialsFileManager() as cred_file:
            return cred_file.exists()

    @property
    def token(self) -> Token:
        with TOKEN_REFRESH_LOCK:
            if not self._token.is_going_to_be_expired():
                return self._token

            logger.info("Refreshing access token")
            with self._cred_file:
                new_cred_file_content = self._cred_file.read()
                new_token = new_cred_file_content.to_token()
                new_host = new_cred_file_content.host

                if new_cred_file_content == self._last_cred_file_content:
                    self._token = self._auth_service.refresh_token(self._token)
                    self._last_cred_file_content = CredentialsFileContent(
                        host=self._host,
                        access_token=self._token.access_token,
                        refresh_token=self._token.refresh_token,
                    )
                    self._cred_file.write(self._last_cred_file_content)
                    return self._token

                if (
                    new_host == self._host
                    and new_token.to_user_info() == self._token.to_user_info()
                ):
                    self._last_cred_file_content = new_cred_file_content
                    self._token = new_token
                    # recursive
                    return self.token

                raise Exception(
                    "Credentials on disk changed while mlfoundry was running."
                )

    @property
    def base_url(self) -> str:
        return self._host



from typing import Dict, List

import numpy as np


class Model:
    def __init__(self, model_path: str):
        self.model_path = model_path

    def predict(self, preprocessed_data: List[List[List[int]]]) -> Dict:
        print(np.asarray(preprocessed_data))
        print(f"prediction running for model_path {self.model_path}")
        return {"prediction": "cat"}



from __future__ import annotations

import abc
import json
from typing import TYPE_CHECKING

import requests

from servicefoundry.function_service.remote.context import QUAL_NAME_TO_URL_MAP

if TYPE_CHECKING:
    from servicefoundry.function_service.remote.remote import Remote


class Method(abc.ABC):
    def __call__(self, *args, **kwargs):
        if len(args) > 0:
            raise Exception(
                "Positional arguments are not supported now.\n"
                "Please use keyword arguments"
            )
        return self.run(**kwargs)

    @abc.abstractmethod
    def run(self, **kwargs):
        ...

    @abc.abstractmethod
    async def run_async(self, **kwargs):
        ...


class LocalMethod(Method):
    def __init__(self, remote_object: Remote, method):
        self._remote_object = remote_object
        self._method_name = method.__name__

    def run(self, **kwargs):
        return getattr(self._remote_object.instance, self._method_name)(**kwargs)

    async def run_async(self, **kwargs):
        raise NotImplementedError()


class RemoteMethod(Method):
    def __init__(self, remote_object: Remote, method):
        self._remote_object = remote_object
        self._method_name = method.__name__
        self._qual_name = self._remote_object.get_qual_name(method)

    def check(self) -> bool:
        return self._qual_name in QUAL_NAME_TO_URL_MAP

    def run(self, **kwargs):
        url = QUAL_NAME_TO_URL_MAP.get(self._qual_name)
        r = requests.post(url, json=kwargs)
        assert r.status_code == 200, r.text
        return json.loads(r.text)

    async def run_async(self, **kwargs):
        raise NotImplementedError()


def method_factory(*args, **kwargs) -> Method:
    remote_method = RemoteMethod(*args, **kwargs)
    if remote_method.check():
        return remote_method

    return LocalMethod(*args, **kwargs)



import logging
import sys

import rich_click as click

from servicefoundry import logger
from servicefoundry.cli.commands import (
    deploy_v2_command,
    get_build_command,
    get_build_logs_command,
    get_create_command,
    get_delete_command,
    get_get_command,
    get_list_command,
    get_login_command,
    get_logout_command,
    get_logs_command,
    get_patch_command,
    get_redeploy_command,
    get_trigger_command,
)
from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.const import GROUP_CLS
from servicefoundry.cli.util import setup_rich_click
from servicefoundry.lib.util import is_debug_env_set, is_experimental_env_set
from servicefoundry.version import __version__


def _add_experimental_commands(cli):
    from servicefoundry.cli.commands.infra_bootstrap import get_infra_command

    cli.add_command(get_infra_command())


def create_servicefoundry_cli():
    """Generates CLI by combining all subcommands into a main CLI and returns in

    Returns:
        function: main CLI functions will all added sub-commands
    """
    cli = servicefoundry_cli
    cli.add_command(get_login_command())
    cli.add_command(get_get_command())
    cli.add_command(get_list_command())
    cli.add_command(get_delete_command())
    cli.add_command(get_create_command())
    cli.add_command(get_redeploy_command())
    cli.add_command(get_logout_command())
    cli.add_command(get_build_command())
    cli.add_command(deploy_v2_command)
    cli.add_command(get_build_logs_command())
    cli.add_command(get_logs_command())
    cli.add_command(get_trigger_command())

    if not (sys.platform.startswith("win32") or sys.platform.startswith("cygwin")):
        cli.add_command(get_patch_command())

    if is_experimental_env_set():
        _add_experimental_commands(cli)
    return cli


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(
    cls=GROUP_CLS, context_settings=CONTEXT_SETTINGS, invoke_without_command=True
)
@click.option(
    "--json",
    is_flag=True,
    help="Output entities in json format instead of formatted tables",
)
@click.option(
    "--debug",
    is_flag=True,
    default=is_debug_env_set,
    help="Set logging level to Debug. Can also be set using environment variable. E.g. SFY_DEBUG=1",
)
@click.version_option(__version__)
@click.pass_context
def servicefoundry_cli(ctx, json, debug):
    """
    Servicefoundry provides an easy way to deploy your services, jobs and models.
    \b

    To start, login to your Truefoundry account with `sfy login`

    Then start deploying with `sfy deploy`

    Deploy a Service: https://docs.truefoundry.com/docs/deploying-your-first-service

    Deploy a Job: https://docs.truefoundry.com/docs/jobs-quickstart

    Deploy a Model: https://docs.truefoundry.com/docs/scikit-learn
    """
    setup_rich_click()
    # TODO (chiragjn): Change this to -o json|yaml|table|pager
    CliConfig.set("json", json)
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
    logger.add_cli_handler(level=logging.DEBUG if debug else logging.INFO)



import json
import os

from servicefoundry.lib.config.config_manager import ConfigManager
from servicefoundry.lib.config.dict_questionaire import DictQuestionaire

# TODO (chiragjn): This file has a bunch of unresolved references and attributes
# TODO: This should just return the input yaml to me


class InstallationConfig:
    def __init__(self, infra_config_file_path=None) -> None:

        if infra_config_file_path and os.path.exists(infra_config_file_path):
            with open(infra_config_file_path, "r") as config:
                infra_config = json.loads(config.read())
        else:
            infra_config = ConfigManager().get_config(ConfigManager.INSTALL_CONFIG)
        if infra_config:
            self.inputs = infra_config["inputs"]
            self.provisioning = infra_config["provisioning"]
            self.ubermold = infra_config["ubermold"]

    # User input needed for installation
    inputs = {
        "provider": None,
        # add all inputs here
    }
    # Config needed for the provisioning step where terraform is run
    provisioning = {
        "provider": None,
        # add all outputs here
    }
    # Config needed for the ubermold step where helm apply is run
    ubermold = {
        # add all outputs here
    }

    def populate_bootstrapping_config(self, target_repo_config, base_terragrunt_dir):
        # TODO: use new output manifest
        print("TODO: write the function")

    def populate_provisioning_config(self, provider: str):
        self.provisioning["provider"] = "aws"
        DictQuestionaire(self.provisioning["awsInputs"]).ask(
            custom_handlers={"externalNetworkConfig": self.__externalConfigHandler}
        )
        ConfigManager().save_config(ConfigManager.INFRA_CONFIG, self.to_json())

    def to_json(self):
        return {"provisioning": self.provisioning, "bootstrapping": self.bootstrapping}



import os
from shutil import which

import questionary

from servicefoundry.lib.binarydownloader import BinaryDependencies, BinaryName
from servicefoundry.lib.clients.shell_client import Shell
from servicefoundry.logger import logger

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def check_or_install_dependencies(dependencies):
    """
    Takes a list of dependencies and validates if they're present in the system
    If not present it tries to install the the binaries into troufoundry bin folder
    """
    for dep in dependencies:
        try:
            if which(dep):
                logger.info(f"{dep} found on local system")
            else:
                BinaryDependencies().which(BinaryName(dep))
                logger.info(f"{dep} installed automatically in .truefoundry/bin")
        except Exception:
            raise Exception(
                f"Automated installation of {dep} not yet supported. Refer to docs for manual installation"
            )


def execute_kubectl_apply_file(file_path):
    return Shell().execute_shell_command(
        [
            BinaryDependencies().which(BinaryName.KUBECTL),
            "apply",
            "-f",
            file_path,
        ]
    )


def set_current_kubecontext():
    """
    This method lists down the current kubecontext available to the user and asks the user
    to select one kubecontext. If no kubecontexts are found, it throws an exception.
    If there is only one kubecontext found, it will select that kubecontext and print
    the message to the user. Once the user selects the kubecontext, it will set the selected
    context as the current context
    """
    kube_contexts = Shell().execute_shell_command(
        [
            BinaryDependencies().which(BinaryName.KUBECTL),
            "config",
            "get-contexts",
            "--no-headers",
            "--output",
            "name",
        ]
    )
    contexts_list = kube_contexts.splitlines()
    selected_context = questionary.select(
        f"Please select the cluster where you want to install Truefoundry: ",
        choices=contexts_list,
    ).ask()
    # Set the selected kubecontext as the current kubecontext
    Shell().execute_shell_command(
        [
            BinaryDependencies().which(BinaryName.KUBECTL),
            "config",
            "use-context",
            selected_context,
        ]
    )
    logger.info(f"Set current context as {selected_context}")
    return selected_context


# This functions checks if the input crd_name is already installed in the cluster


def check_if_crd_installed(name, crd_name) -> bool:
    try:
        check_crd = Shell().execute_shell_command(
            [
                BinaryDependencies().which(BinaryName.KUBECTL),
                "get",
                "crd",
                crd_name,
                "-ojson",
            ]
        )
        return True
    except Exception:
        logger.info(f"{name} not installed on cluster. Moving ahead...")
        return False


def check_if_truefoundry_exists():
    # check if truefoundry namespace is present
    try:
        Shell().execute_shell_command(
            [
                BinaryDependencies().which(BinaryName.KUBECTL),
                "get",
                "secrets",
                "servicefoundry-server-env-secret",
                "-n",
                "truefoundry",
            ]
        )
        return True
    except Exception:
        logger.info("Truefoundry not installed on the cluster")
        return False


def create_namespace_if_not_exists(namespace_name: str):
    try:
        namespace = Shell().execute_shell_command(
            [
                BinaryDependencies().which(BinaryName.KUBECTL),
                "create",
                "namespace",
                namespace_name,
            ]
        )
    except Exception as e:
        logger.info(e)
        return


def execute_nats_bootstrap_script():
    if os.path.exists("./nsc"):
        raise Exception(
            "nsc directory already exists. Please delete that before proceeding"
        )
    Shell().execute_shell_command(
        ["/bin/bash", os.path.join(THIS_DIR, "nats-bootstrap.sh")]
    )
    nats_controlplane_account_seed = ""
    with open("./nsc/tfy.seed", "r") as f:
        nats_controlplane_account_seed = f.readline()
    return nats_controlplane_account_seed


def delete_directory(dir):
    try:
        Shell().execute_shell_command(["rm", "-rf", dir])
    except Exception as e:
        logger.info(e)
        return


def install_argocd_chart(argocd_chart_version):
    logger.info("Installing ArgoCD")
    try:
        print(
            Shell().execute_shell_command(
                [
                    which("helm"),
                    "upgrade",
                    "--install",
                    "--namespace",
                    "argocd",
                    "--create-namespace",
                    "--repo",
                    "https://argoproj.github.io/argo-helm",
                    "argocd",
                    "argo-cd",
                    "--set",
                    "server.extraArgs={--insecure},applicationSet.enabled=false,notifications.enabled=false,dex.enabled=false",
                    "--version",
                    argocd_chart_version,
                ]
            )
        )
        logger.info("ArgoCD Installation Done")
    except Exception as e:
        raise Exception("Error installing ArgoCD helm chart. Please try again")


def install_ubermold_helm(helm_local_path):
    if not questionary.confirm(
        "Do you want to continue with istio, prometheus and truefoundry installation?",
        default=True,
    ).ask():
        return
    logger.info("Installing Istio, prometheus and truefoundry components")
    try:
        print(
            Shell().execute_shell_command(
                [
                    which("helm"),
                    "upgrade",
                    "--install",
                    "ubermold",
                    helm_local_path,
                    "-n",
                    "argocd",
                ]
            )
        )
        logger.info("ArgoCD Installation Done")
    except Exception as e:
        raise Exception("Error installing helm chart. Please try again.")



from typing import Optional

from servicefoundry.cli.console import console
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.messages import (
    PROMPT_CREATING_NEW_WORKSPACE,
    PROMPT_USING_CLUSTER_CONTEXT,
)
from servicefoundry.lib.model.entity import Workspace, WorkspaceResources
from servicefoundry.logger import logger


def create_workspace(
    name: str,
    cluster_name: str,
    cpu_limit: Optional[float] = None,
    memory_limit: Optional[int] = None,
    ephemeral_storage_limit: Optional[int] = None,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()

    workspace_resources = WorkspaceResources(
        cpu_limit=cpu_limit,
        memory_limit=memory_limit,
        ephemeral_storage_limit=ephemeral_storage_limit,
    )

    console.print(PROMPT_USING_CLUSTER_CONTEXT.format(cluster_name))
    with console.status(PROMPT_CREATING_NEW_WORKSPACE.format(name), spinner="dots"):
        workspace = client.create_workspace(
            workspace_name=name,
            cluster_name=cluster_name,
            resources=workspace_resources,
        )

    url = f"{client.base_url.strip('/')}/workspaces"
    logger.info(
        "You can find your workspace: '%s' on the dashboard: %s", workspace.name, url
    )

    return workspace


def list_workspaces(
    cluster_name: Optional[str] = None,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    workspaces = client.list_workspaces(cluster_id=cluster_name)
    return workspaces


def get_workspace_by_fqn(
    workspace_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()
    workspaces = client.get_workspace_by_fqn(workspace_fqn=workspace_fqn)
    if len(workspaces) == 0:
        raise ValueError(f"Workspace with FQN {workspace_fqn!r} does not exist.")
    workspace = workspaces[0]
    return workspace


def delete_workspace(
    workspace_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()
    workspace = client.get_id_from_fqn(fqn=workspace_fqn, fqn_type="workspace")
    deleted_workspace = client.remove_workspace(workspace_id=workspace["workspaceId"])
    console.print(f"[yellow]Deleted workspace {deleted_workspace.fqn!r}[/]")
    return deleted_workspace



import logging

from my_module import normal, uniform

from servicefoundry import Resources
from servicefoundry.function_service import BuildConfig, FunctionService

logging.basicConfig(level=logging.INFO)

random_service = FunctionService(
    name="random-service-2",
    build_config=BuildConfig(pip_packages=["numpy<1.22.0"]),
    resources=Resources(cpu_limit=0.5, memory_limit=512),
    replicas=2,
    port=4000,
)
random_service.register_function(normal)
random_service.register_function(uniform)

print(random_service)
# random_service.deploy("v1:local:my-ws-2")
random_service.run().join()
